#!/usr/bin/env python

# -----------------------------------------------------------------------------
# HMM helper functions
# Author: Mary Richardson
# Date: 2020.01.08
# -----------------------------------------------------------------------------

import logging
import numpy as np
from scipy.special import logsumexp

logger = logging.getLogger(__name__)

# Ignore divide by zero error for log
np.seterr(divide='ignore')


def logdotexp(A, B):
    """
    Code inspiration: https://stackoverflow.com/questions/23630277/numerically-stable-way-to-multiply-log-probability-matrices-in-numpy
    Given matrices A (one-hot encoded) and B (log space), calculate the
    dot product computed in logspace using the logsumexp function such that
    the log values of A are multiplied by B and then summed (in linear space)
    Args:
        A: numpy array (one-hot encoded)
        B: numpy array (log space)
    Returns:
        C: numpy array resulting from the dot product of A and B

    """
    max_B = np.max(A)
    C = np.log(np.dot(A, np.exp(B - max_B)))
    C += max_B

    return C


def one_hot_nt(x, nts):
    """
    Generate the one-hot encoding for this nucleotide sequence
    Args:
        x: the nucleotide sequence
        nts: list of valid nucleotides
    Returns:
        one_hot_xt: the one-hot-encoded matrix for this nucleotide sequence
    Raises:
        ValueError: Invalid nucleotide sequence
    """
    # Length of sequence
    L = len(x)

    # Number of nucleotides
    N = len(nts)

    # Convert the sequence to numbers (A=0, C=1, G=2, T=3)
    one_hot_x = np.zeros((L,N))
    for i,nt in enumerate(x):
        if nt in nts:
            one_hot_x[i, nts.index(nt)] = 1
        elif nt=='N': # Any nucleotide
            one_hot_x[i, :] = np.ones(N)
        elif nt=='R': # Purine
            one_hot_x[i, nts.index('A')] = 1
            one_hot_x[i, nts.index('G')] = 1
        elif nt=='Y': # Pyrimidine
            one_hot_x[i, nts.index('C')] = 1
            one_hot_x[i, nts.index('T')] = 1
        else:
            logging.error(('Invalid nucleotide sequence. '
                           'Unexpected character in sequence: %s' % i))

    return one_hot_x


def one_hot_riboseq(r, rho_ranges):
    """
    Generate the one-hot encoding for this rho sequence
    Args:
        r: the rho sequence
        rho_ranges: array of riboseq emission bin ranges
    Returns:
        one_hot_rho: the one-hot-encoded matrix for this rho sequence
    Raises:
        ValueError: Invalid ribosome profiling values
    """
    # Length of sequence
    L = len(r)

    # Number of discretized riboseq ranges
    R = len(rho_ranges)-1

    try:
        # Convert the riboseq sequence to discretized ranges
        # Include the upper limit of the range and shift to be zero-indexed
        y = np.digitize(r, rho_ranges, right=True) - 1

        # For anything beyond the max observed value,
        # assign it to the uppermost bin
        y[y >= R] = R-1

        # Create a one-hot encoding for the sequence
        one_hot_r = np.eye(R)[np.array(y)]

    except:
        logging.error('Invalid ribosome profiling values.')

    return one_hot_r


def convert_emissions(x, r, log_emit_nt, log_emit_rho, nts, rho_ranges):
    """
    Generate the log emission probabilities at each position of a sequence
    Note: Returns the emission probabilities at each sequence position as
    an L x M matrix. This allows matrix addition down the line, since the
    forward and backward matrices are L x M
    Args:
        x: the nucleotide sequence
        r: the rho sequence
        log_emit_nt: log nucleotide emission probabilities
        log_emit_rho: log riboseq emission probabilities
        nts: list of valid nucleotides
        rho_ranges: array of riboseq emission bin ranges
    Returns:
        one_hot_x: the one-hot-encoded matrix for this nucleotide sequence
        one_hot_r: the one-hot-encoded matrix for this rho sequence
        log_emit: the total emission probabilities at each position for
            this sequence (combined nucleotide and rho emissions)
    """
    # Calculate the nucleotide emissions for this sequence
    one_hot_x = one_hot_nt(x, nts)
    log_emit_x = logdotexp(one_hot_x, log_emit_nt.T)

    # Create a one-hot encoding for this riboseq sequence
    one_hot_r = one_hot_riboseq(r, rho_ranges)
    log_emit_r = logdotexp(one_hot_r, log_emit_rho.T)

    # Calculate the total emissions for this sequence
    log_emit = log_emit_x + log_emit_r

    return log_emit



#!/bin/env python

# -----------------------------------------------------------------------------
# HMM run model functions
# Author: Mary Richardson
# Date: 2021.10.25
# -----------------------------------------------------------------------------

import logging
import numpy as np
import multiprocessing as mp

logger = logging.getLogger(__name__)

from orfeus.hmm.helper import convert_emissions
from orfeus.hmm.algorithms import viterbi, score_path

NUCLEOTIDES = ['A','C','G','T']


def get_annotated_path(path, states):
    """
    Convert a path of states to a path of state indices
    Args:
        path: path of states
        states: list of state names
    Returns:
        path: path of state indices
    Raises:

    """
    # Get the annotated state path
    try:
        path = [states.index(s) for s in path]
    except:
        path = []

    return path


def get_transcripts(data_df,
                    states,
                    coverage=0):
    """
    Generates a list of transcripts with at least the specified mean coverage
    Args:
        data_df: pandas DataFrame with all transcript info
        states: list of state names
        coverage: (optional) minimum mean coverage threhold
    Returns:
        transcripts: dict of transcripts with at least the minimum coverage
    Raises:

    """
    # Get the transcript properties of interest
    df = data_df[['transcript_id',
                  'transcript_name',
                  'seq',
                  'log_reads',
                  'state',
                  'mean_reads']]

    # Filter down to transcripts above the coverage threshold
    df = df[df['mean_reads'] > coverage]

    # Group by transcript
    df['transcript_id'] = df['transcript_id'].apply(str)
    df['transcript_name'] = df['transcript_name'].apply(str)
    df['mean_reads'] = df['mean_reads'].apply(float)

    df = df.groupby(['transcript_id',
                     'transcript_name',
                     'mean_reads'],
                      as_index=False) \
           .agg({'seq': lambda x: list(x),
                 'log_reads': lambda x: list(x),
                 'state': lambda x: get_annotated_path(list(x), states)})

    # Save transcript
    transcripts = df.values.tolist()

    return transcripts


def hmm_single(x, r,
               path,
               parameters_h1,
               parameters_h0):
    """
    Run the mdoel on a single transcript
    Args:
        x: nucleotide sequence
        r: rho sequence
        path: annotated path
        parameters_h1: list of all parameter matrices for the current model
        parameters_h0: list of all parameter matrices for the null model
    Returns:
        result: list of results of running the model on this transcript
    Raises:

    """
    # Unpack the arguments
    states_h1, states_idx_h1, altorfs_idx_h1, \
    p_emit_nt_h1, p_emit_rho_h1, rho_ranges_h1, p_trans_h1, p_trans_rev_h1, \
    prev_states_options_h1, next_states_options_h1 = parameters_h1

    states_h0, states_idx_h0, altorfs_idx_h0, \
    p_emit_nt_h0, p_emit_rho_h0, rho_ranges_h0, p_trans_h0, p_trans_rev_h0, \
    prev_states_options_h0, next_states_options_h0 = parameters_h0

    log_emit_nt_h1 = np.log(p_emit_nt_h1)
    log_emit_rho_h1 = np.log(p_emit_rho_h1)
    log_emit_nt_h0 = np.log(p_emit_nt_h0)
    log_emit_rho_h0 = np.log(p_emit_rho_h0)
    log_trans_h1 = np.log(p_trans_h1)
    log_trans_h0 = np.log(p_trans_h0)

    # Generate the one-hot encoding and reformatted emission probabilities
    # for this sequence
    log_emit_h1 = convert_emissions(x,
                                    r,
                                    log_emit_nt_h1,
                                    log_emit_rho_h1,
                                    NUCLEOTIDES,
                                    rho_ranges_h1)
    log_emit_h0 = convert_emissions(x,
                                    r,
                                    log_emit_nt_h0,
                                    log_emit_rho_h0,
                                    NUCLEOTIDES,
                                    rho_ranges_h0)

    # Predict the viterbi state path
    path_h1, P_h1 = viterbi(log_emit_h1,
                            log_trans_h1,
                            prev_states_options_h1,
                            next_states_options_h1)
    path_h0, P_h0 = viterbi(log_emit_h0,
                            log_trans_h0,
                            prev_states_options_h0,
                            next_states_options_h0)

    # Calculate the probability of the annotated (h2), predicted (h1), and
    # null (h0) paths
    path_score_h2 = score_path(path, log_emit_h1, log_trans_h1)
    path_score_h1 = score_path(path_h1, log_emit_h1, log_trans_h1)
    path_score_h0 = score_path(path_h0, log_emit_h1, log_trans_h1)

    # Calculate the log-odds score
    log_odds_score = path_score_h1 - path_score_h0

    # Make a list of all values to return for this transcript
    return path, path_h1, path_h0, \
           path_score_h2, path_score_h1, path_score_h0, log_odds_score, \
           log_emit_h1, log_trans_h1


def hmm_parallel(args):
    """
    Run the mdoel on a single transcript
    Args: (all packaged into the list args)
        transcripts: dict of transcripts
        parameters_h1: list of all parameter matrices for the current model
        parameters_h0: list of all parameter matrices for the null model
    Returns:
        result: list of results of running the model on this transcript
    Raises:

    """
    # Unpack the arguments
    transcript, parameters_h1, parameters_h0 = args

    transcript_id, transcript_name, coverage, x, r, path = transcript

    path_h2, path_h1, path_h0, \
    path_score_h2, path_score_h1, path_score_h0, log_odds_score, \
    log_emit_h1, log_trans_h1 = hmm_single(x, r,
                                           path,
                                           parameters_h1,
                                           parameters_h0)

    # Make a list of all values to return for this transcript
    result = [transcript_id, transcript_name, coverage, x, r, \
              path_h2, path_h1, path_h0, \
              path_score_h2, path_score_h1, path_score_h0, \
              log_odds_score]

    logging.info('%s (%s)' % (transcript_id,transcript_name))

    return result


def run_hmm(data_df,
            coverage,
            parameters_h1,
            parameters_h0):
    """
    Run the model on multiple transcripts in parallel
    Args:
        data_df: processed data
        coverage: mean reads per transcript coverage threshold
        parameters_h1: list of all parameter matrices for the current model
        parameters_h0: list of all parameter matrices for the null model
    Returns:
        results: array containing all results of running the model
    Raises:

    """
    # Unpack the parameters
    states, states_idx, altorfs_idx, \
    p_emit_nt, p_emit_rho, rho_ranges, p_trans, p_trans_rev, \
    prev_states_options, next_states_options = parameters_h1

    # Get transcripts
    transcripts = get_transcripts(data_df,
                                  states,
                                  coverage)

    with mp.Pool(5) as pool:

        # Number of sequences
        C = len(transcripts)

        # For each sequence, calculate the predicted viterbi path
        args = [(transcripts[c], parameters_h1, parameters_h0) \
                 for c in range(C)]
        results_paths = pool.map(hmm_parallel, args)

    pool.join()

    # Get results as an array of prediction dictionaries
    results = np.array(results_paths)

    return results



#!/usr/bin/env python

# HMM tests
# Author: Mary Richardson
# Date: 2020.01.08

import sys
import argparse
import numpy as np
import pandas as pd
import pickle as pk
import itertools as it
from scipy.special import logsumexp
import matplotlib.pyplot as plt
from matplotlib.ticker import FormatStrFormatter as fmt
import seaborn as sns

from helper import convert_emissions
from algorithms import viterbi, forward, backward, score_path
from set_parameters import riboseq_emission_probs

# Customize plotting style
sns.set_style('white')
colors = ['#CA4B4C', '#F0BF1A', '#2E659D']
sns.set_palette(sns.color_palette(colors))

np.set_printoptions(threshold=sys.maxsize)



###########################
# Check parameters
###########################


def check_sums(p_emit, p_trans):
    '''
    Check that all probabilities sum to 1 where they should
    '''
    match = True # Default

    # Check that emission prob matrix rows except begin state sum to 1 (within some error here)
    if not (np.all(p_emit.sum(axis=1) > 0.99) and (np.all(p_emit.sum(axis=1) < 1.01))):
        print('Emission probability matrix rows don\'t sum to 1!')
        match = False

    # Check that transition prob matrix rows sum to 1
    if (np.all(p_trans.sum(axis=1) != 1)):
        print('Transition probability matrix rows don\'t sum to 1!')
        match = False

    return match



###########################
# Check HMM functions
###########################

def check_scores(log_emit_nx, log_trans,
                 prev_states_options, next_states_options
                 dec=5):
    '''
    Given an observed sequence x and parameters,
    Checks that at every column the product of the forward and backward algorithms summed over all states is equal
    to the total forward score and total backward score
    '''
    L = len(log_emit_nx)  # Length of observed sequence

    # Calculate the forward and backward scores at each position of the sequence
    F, F_score = forward(log_emit_nx, log_trans, prev_states_options, next_states_options)
    B, B_score = backward(log_emit_nx, log_trans, prev_states_options, next_states_options)

    # Check whether forward score equals backward score to start
    match = True # Default
    if (np.round(F_score, dec) != np.round(B_score, dec)):
        print('Forward and backward scores not equal!')
        match = False

    # Calculate the score at each position of the sequence
    for i in range(L):
        sequence_score = logsumexp(F[i,:] + B[i,:])
        if np.round(sequence_score, dec) != np.round(F_score, dec):
            print('The scores at each position of the sequence are not equal!')
            match = False

    # Return whether the forward score, backward score, and the score at each position match
    return match


def brute_test(log_emit_nx, log_trans,
               prev_states_options, next_states_options,
               dec=5):
    '''
    Given an observed sequence x and parameters,
    Enumerates all possible paths and checks that the sum over all paths is equal to the total
    forward score
    '''
    M = len(log_trans) - 1 # Number of states
    L = len(log_emit_nx)   # Length of observed sequence

    # Enumerate all possible hidden state paths for this sequence (M^L paths)
    # Compute the cartesian product of the possible states at each position in the sequence
    all_paths = list(it.product(range(M), repeat = L))

    # Compute the probability of each possible path
    path_scores = []
    for path in all_paths:
        path_score = score_path(path, log_emit_nx, log_trans)
        path_scores.append(path_score)

    # Sum the probability over all paths
    all_paths_score = logsumexp(path_scores)
    print('Sum over all paths:  ', all_paths_score)

    # Calculate the total score from the forward algorithm
    F, F_score = forward(log_emit_nx, log_trans, prev_states_options, next_states_options)
    print('Total forward score: ', F_score)

    # Check that the sum over all paths equals the total forward probability
    match = True # Default
    if (np.round(F_score, dec) != np.round(all_paths_score, dec)):
        print('The sum over all paths and the forward score are not equal!')
        match = False

    return match


def generation_test(log_emit_nt, log_emit, log_trans,
                    prev_states_options, next_states_options,
                    nts, riboseq_ranges, dec=5, suppress=True):
    '''
    Given parameters,
    Samples a sequence and calculates the viterbi path to check that the viterbi result
    is at least as probable as the true path but not greater than the forward and backward scores
    '''
    # Simulate a sequence
    n, x, true_path = simulate_seq(np.exp(log_emit_nt), np.exp(log_emit), np.exp(log_trans), nts, riboseq_ranges)

    # Generate the reformatted emission probabilities
    one_hot_n, one_hot_x, log_emit_nx = convert_emissions(n, x, log_emit_nt, log_emit, nts, riboseq_ranges)

    # Calculate the viterbi path
    viterbi_path, P = viterbi(log_emit_nx, log_trans, prev_states_options, next_states_options)

    # Calculate the score of the true path
    true_score = score_path(true_path, log_emit_nx, log_trans)
    print('True path score:      ', true_score)

    # Calculate the viterbi path score
    viterbi_score = score_path(viterbi_path, log_emit_nx, log_trans)
    print('Viterbi path score:   ', viterbi_score)

    # Calculate the total score from the forward algorithm
    F, F_score = forward(log_emit_nx, log_trans, prev_states_options, next_states_options)
    print('Total forward score:  ', F_score)

    # Calculate the total score from the backward algorithm
    B, B_score = backward(log_emit_nx, log_trans, prev_states_options, next_states_options)
    print('Total backward score: ', B_score)

    # Check that the forward and backward scores are equal
    match = True # Default
    if not np.round(F_score, dec) == np.round(B_score, dec):
        print('Forward and backward scores not equal!')
        match = False

    # Check that the forward and backward scores are at least as good as the viterbi score
    if not F_score >= viterbi_score and B_score >= viterbi_score:
        print('Forward and backward scores not at least as good as viterbi!')
        match = False

    # Check that the viterbi path is at least as probable as the actual path
    if not viterbi_score >= true_score:
        print('Viterbi score not at least as good as true score!')
        match = False

    return match


def main():

    # Accept command line arguments for input and output files
    parser = argparse.ArgumentParser(description='Test the model for the current dataset')

    parser.add_argument('nt_emissions_file', type=str,
                        help='nt emission probability file path')
    parser.add_argument('riboseq_emissions_file', type=str,
                        help='riboseq emission probability file path')
    parser.add_argument('transitions_file', type=str,
                        help='transition probability file path')
    parser.add_argument('transitions_rev_file', type=str,
                        help='reverse transition probability file path')
    parser.add_argument('constants_file', type=str,
                        help='constants file path')

    args = parser.parse_args()


    ###########################
    # Import initial transition and emission probabilities
    ###########################

    # Get the constants
    [nts, riboseq_ranges, states] = pk.load(open(args.constants_file, 'rb'))
    M = len(states)
    R = len(riboseq_ranges)-1

    # Get the initial parameters
    p_emit_nt = np.loadtxt(args.nt_emissions_file)
    p_emit = np.loadtxt(args.riboseq_emissions_file)
    p_trans = np.loadtxt(args.transitions_file)
    p_trans_rev = np.loadtxt(args.transitions_rev_file)

    # Log the initial parameters
    log_emit_nt = np.log(p_emit_nt)
    log_emit = np.log(p_emit)
    log_trans = np.log(p_trans)


    ###########################
    # Generate a test sequence
    ###########################

    # Simple test sequence (A=0, C=1, G=2, T=3)
    n = np.array(['A','T','G','T','T'])
    x = np.array([0,1,2,3,1])
    one_hot_n, one_hot_x, log_emit_nx = convert_emissions(n, x, log_emit_nt, log_emit, nts, riboseq_ranges)


    ###########################
    # Check parameters
    ###########################

    print('Check parameters:')
    print(check_sums(p_emit, p_trans))
    print()


    ###########################
    # Check scores
    ###########################

    # Test on simple case
    print('Check scores (simple sequence):')
    print(check_scores(log_emit_nx, log_trans))
    print()

    # Test on random sequence
    print('Check scores (simulated sequence):')
    for i in range(10):
        simulated_n, simulated_x, true_path = simulate_seq(p_emit_nt, p_emit, p_trans, nts, riboseq_ranges)
        one_hot_n, one_hot_x, simulated_log_emit_nx = convert_emissions(simulated_n, simulated_x, log_emit_nt, log_emit,
                                                            nts, riboseq_ranges)
        print(check_scores(simulated_log_emit_nx, log_trans))
    print()


    ###########################
    # Brute test
    ###########################

    # Test on simple case
    #print('Brute test (simple sequence):')
    #print(brute_test(log_emit_nx, log_trans))
    #print()

    # Test on random sequence (takes a LONG time)
    #simulated_n, simulated_x, true_path = simulate_seq(p_emit_nt, p_emit, p_trans, nts, riboseq_ranges)
    #one_hot_n, one_hot_x, simulated_log_emit_nx = convert_emissions(simulated_n, simulated_x, log_emit_nt, log_emit,
    #                                                    nts, riboseq_ranges)
    #print('Brute test (simulated sequence):')
    #print(brute_test(simulated_log_emit_nx, log_trans))
    #print()


    ###########################
    # Generation test
    ###########################

    print('Generation test:')
    for i in range(10):
        print(generation_test(log_emit_nt, log_emit, log_trans, nts, riboseq_ranges))
    print()


    ###########################
    # Baum-Welch test
    ###########################

    # Generate synthetic training data from known parameters
    train_id, train_n, train_x = [], [], []
    for i in range(1000):

        # Simulate a sequence
        simulated_n, simulated_x, true_path = simulate_seq(p_emit_nt, p_emit, p_trans, nts, riboseq_ranges)
        predict_frameshift(true_path, states)

        # Add it to the training data
        train_id.append('simulated %i' % i)
        train_n.append(simulated_n)
        train_x.append(simulated_x)

    # Save the simulated transcripts
    transcripts_file = 'tmp1.txt'
    pk.dump([train_id, train_n, train_x], open(transcripts_file, 'wb'))


    # Start from true parameters
    print('Baum-Welch test (starting from true parameters):')
    log_emit_nt_test = log_emit_nt
    log_emit_test  = log_emit
    log_trans_test = log_trans
    match, log_emit_nt_test, log_emit_test, log_trans_test, log_emit_nt_pred, log_emit_pred, log_trans_pred, ll  = test_baum_welch(transcripts_file, log_emit_nt, log_emit, log_trans,
                                 log_emit_nt_test, log_emit_test, log_trans_test,
                                 nts, riboseq_ranges)
    print_baum_welch(states,
           log_emit_nt, log_emit, log_trans,
           log_emit_nt_test, log_emit_test, log_trans_test,
           log_emit_nt_pred, log_emit_pred, log_trans_pred)
    print('final log likelihood: %f' % ll)
    print(match)
    print()


    # Set random parameters to start
    print('Baum-Welch test (starting from random parameters):')
    best_ll = -np.inf
    for _ in range(50): # Run from multiple random starting conditions

        p_emit_test = np.zeros((M,R))
        p_emit_test[np.nonzero(p_emit)] = 1
        for m in range(M):
            nonzero = np.transpose(p_emit_test[m].nonzero()).flatten()
            p = np.random.dirichlet(np.ones(len(nonzero)))
            for i,j in enumerate(nonzero):
                p_emit_test[m,j] = p[i]

        p_trans_test = np.zeros((M+1,M+1))
        p_trans_test[np.nonzero(p_trans)] = 1
        for m in range(M):
            nonzero = np.transpose(p_trans_test[m].nonzero()).flatten()
            p = np.random.dirichlet(np.ones(len(nonzero)))
            for i,j in enumerate(nonzero):
                p_trans_test[m,j] = p[i]

        log_emit_test = np.log(p_emit_test)
        log_trans_test = np.log(p_trans_test)

        match, log_emit_nt_test, log_emit_test, log_trans_test, log_emit_nt_pred, log_emit_pred, log_trans_pred, ll  = test_baum_welch(transcripts_file, log_emit_nt, log_emit, log_trans,
                                     log_emit_nt_test, log_emit_test, log_trans_test,
                                     nts, riboseq_ranges)


        if ll > best_ll:
            best_match = match
            best_log_emit_nt_test = log_emit_nt_test
            best_log_emit_test = log_emit_test
            best_log_trans_test = log_trans_test
            best_log_emit_nt_pred = log_emit_nt_pred
            best_log_emit_pred = log_emit_pred
            best_log_trans_pred = log_trans_pred
            best_ll = ll


    print_baum_welch(states,
           log_emit_nt, log_emit, log_trans,
           best_log_emit_nt_test, best_log_emit_test, best_log_trans_test,
           best_log_emit_nt_pred, best_log_emit_pred, best_log_trans_pred)
    print('final log likelihood: %f' % best_ll)
    print(best_match)
    print()



if __name__== "__main__":
    main()



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# Build ORFeus model
# Author: Mary Richardson
# Date: 2023.03.03
# -----------------------------------------------------------------------------

import argparse
import logging
import os
import numpy as np

from orfeus.parsers.data import read_data_file
from orfeus.hmm.run import run_hmm
from orfeus.output.text import write_results_fasta, write_results_table
from orfeus.output.logo import print_logo


def main():

# ------------------------------------------------------------------------------
# Accept command line arguments
# ------------------------------------------------------------------------------

    parser = argparse.ArgumentParser(description='Run ORFeus model')

    # Input data
    parser.add_argument('data_file', type=str,
                        help=('processed data file path '
                            '(generated by orfeus_build_model.py)'))
    parser.add_argument('parameters_file_h1', type=str,
                        help=('model parameters file path '
                            '(generated by orfeus_build_model.py)'))
    parser.add_argument('parameters_file_h0', type=str,
                        help=('null model parameters file path '
                            '(generated by orfeus_build_model.py)'))

    # ORFeus options
    parser.add_argument('-o', '--outdir', type=str, default=os.getcwd(),
                        help=('output directory for all intermediate and '
                        'results files'))
    parser.add_argument('-c', '--coverage', type=float, default=0.1,
                        help=('mean riboseq coverage threshold'))

    args = parser.parse_args()

    print_logo()

# ------------------------------------------------------------------------------
# Prepare output directory
# ------------------------------------------------------------------------------

    # Create output directory
    if args.outdir == os.getcwd():
        outdir = os.path.join(args.outdir, r'orfeus')
        os.makedirs(outdir, exist_ok=True)
    else:
        if not os.path.isdir(args.outdir):
            print('Output directory does not exist. Please provide a valid path.')
        outdir = args.outdir

    # Set up log file
    #log_file = os.path.join(outdir, 'orfeus_build.log')
    logging.basicConfig(format='%(levelname)s: %(message)s',
                        filemode = 'w',
                        level = logging.DEBUG)

# ------------------------------------------------------------------------------
# Import data and parameters
# ------------------------------------------------------------------------------

    logging.info('Importing the processed data...')

    # Import the data (specify dtypes where necessary)
    data_df = read_data_file(args.data_file)

    logging.info('Finished importing the data.')


    logging.info('Importing the parameters...')

    parameters_h1 = np.load(args.parameters_file_h1, allow_pickle=True)
    parameters_h0 = np.load(args.parameters_file_h0, allow_pickle=True)

    logging.info('Finished importing the parameters.')

# ------------------------------------------------------------------------------
# Run model
# ------------------------------------------------------------------------------

    logging.info('Running the model...')

    # Calculate the predicted viterbi path for each sequence
    results = run_hmm(data_df,
                      args.coverage,
                      parameters_h1,
                      parameters_h0)

    # Write results to files
    write_results_fasta(results,
                        parameters_h1,
                        os.path.join(outdir, 'predicted_mRNA.fn'),
                        os.path.join(outdir, 'predicted_protein.fa'))

    write_results_table(results,
                        parameters_h1,
                        os.path.join(outdir, 'results.txt'))

    logging.info('Finished running the model.')


if __name__== "__main__":
    main()



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# Set model states
# Author: Mary Richardson
# Date: 2020.07.07, last update 2023.03.02
# -----------------------------------------------------------------------------

import numpy as np


def set_states(start_codon_freqs,
               sense_codon_freqs,
               stop_codon_freqs,
               parameters):
    """
    Generate a list of all states for this model
    Args:
        start_codon_freqs: dict mapping each start codon to its frequency
        stop_codons_freqs: dict mapping each stop codon to its frequency
        sense_codons_freqs: dict mapping each sense codon to its frequency
        parameters: list of parameter values for
            alpha, beta, gamma, delta, zeta
    Returns:
        states: list of state names
        states_idx: dictionary mapping state type to
            state indices (1-indexed)
        altorfs_idx: dictionary mapping more general state type to
            state indices (0-indexed)
    Raises:
        ValueError: Invalid parameter values. Make sure all parameter values
        are greater than or equal to zero.
    """
    # Unpack parameters
    [alpha, beta, gamma, delta, zeta] = parameters

    # List states and their corresponding indices in the
    # transition and emission matrices
    states = []
    states_idx = {}

    # Create dictionary to store all indices for each type of state
    altorfs_idx = {'5UTR':[],
                   '3UTR':[],
                   'ORF':[],
                   'start':[],
                   'sense':[],
                   'stop':[],
                   'X1':[],
                   'X2':[],
                   'PRF':[],
                   'SCR':[],
                   'uORF':[],
                   'dORF':[],
                   'oORF':[],
                   'iORF':[]}

    # Begin counting indices
    idx = 0

    # Add UTR states
    for state in ['5UTR','3UTR']:
        states += [state]
        states_idx[state] = idx
        altorfs_idx[state] = idx
        idx += 1

    # Add start codon states
    start_states = [(codon + '#') for codon in sorted(start_codon_freqs)]
    states += start_states
    n = len(start_states)
    states_idx['start'] = list(range(idx, idx+n))
    altorfs_idx['start'] = list(range(idx, idx+n))
    altorfs_idx['ORF'] = list(range(idx, idx+n))
    idx += n

    # Add sense codon states
    sense_states = [(codon) for codon in sorted(sense_codon_freqs)]
    states += sense_states
    n = len(sense_states)
    states_idx['sense'] = list(range(idx, idx+n))
    altorfs_idx['sense'] = list(range(idx, idx+n))
    altorfs_idx['ORF'] += list(range(idx, idx+n))
    idx += n

    # Add stop codon states
    stop_states = [(codon + '*') for codon in sorted(stop_codon_freqs)]
    states += stop_states
    n = len(stop_states)
    states_idx['stop'] = list(range(idx, idx+n))
    altorfs_idx['stop'] = list(range(idx, idx+n))
    altorfs_idx['ORF'] += list(range(idx, idx+n))
    idx += n

    # Add PRF states
    if alpha > 0:

        # Add X1 state
        states += ['X1']
        states_idx['X1'] = idx
        altorfs_idx['X1'] = idx
        altorfs_idx['PRF'] = [idx]
        idx += 1

    if alpha > 0 and beta > 0:

        # Add X2 state
        states += ['X2']
        states_idx['X2'] = idx
        altorfs_idx['X2'] = idx
        altorfs_idx['PRF'] += [idx]
        idx += 1

    # Add SCR states
    if gamma > 0:

        # Add stop codon readthrough states
        stop_states = [(codon + '**') for codon in sorted(stop_codon_freqs)]
        states += stop_states
        n = len(stop_states)
        states_idx['stop_SCR'] = list(range(idx, idx+n))
        altorfs_idx['SCR'] = list(range(idx, idx+n))
        idx += n

    # Add uORF and dORF states
    if delta > 0:

        # Add uORF start codon states
        start_states = [(codon + '#u') for codon in sorted(start_codon_freqs)]
        states += start_states
        n = len(start_states)
        states_idx['start_uORF'] = list(range(idx, idx+n))
        altorfs_idx['start_uORF'] = list(range(idx, idx+n))
        altorfs_idx['uORF'] = list(range(idx, idx+n))
        idx += n

        # Add uORF sense codon states
        sense_states = [(codon + 'u') for codon in sorted(sense_codon_freqs)]
        states += sense_states
        n = len(sense_states)
        states_idx['sense_uORF'] = list(range(idx, idx+n))
        altorfs_idx['uORF'] += list(range(idx, idx+n))
        idx += n

        # Add uORF stop codon states
        stop_states = [(codon + '*u') for codon in sorted(stop_codon_freqs)]
        states += stop_states
        n = len(stop_states)
        states_idx['stop_uORF'] = list(range(idx, idx+n))
        altorfs_idx['stop_uORF'] = list(range(idx, idx+n))
        altorfs_idx['uORF'] += list(range(idx, idx+n))
        idx += n

        # Add dORF start codon states
        start_states = [(codon + '#d') for codon in sorted(start_codon_freqs)]
        states += start_states
        n = len(start_states)
        states_idx['start_dORF'] = list(range(idx, idx+n))
        altorfs_idx['start_dORF'] = list(range(idx, idx+n))
        altorfs_idx['dORF'] = list(range(idx, idx+n))
        idx += n

        # Add dORF sense codon states
        sense_states = [(codon + 'd') for codon in sorted(sense_codon_freqs)]
        states += sense_states
        n = len(sense_states)
        states_idx['sense_dORF'] = list(range(idx, idx+n))
        altorfs_idx['dORF'] += list(range(idx, idx+n))
        idx += n

        # Add dORF stop codon states
        stop_states = [(codon + '*d') for codon in sorted(stop_codon_freqs)]
        states += stop_states
        n = len(stop_states)
        states_idx['stop_dORF'] = list(range(idx, idx+n))
        altorfs_idx['stop_dORF'] = list(range(idx, idx+n))
        altorfs_idx['dORF'] += list(range(idx, idx+n))
        idx += n

    # Add oORF states
    if epsilon > 0:

        states_idx['sense_start_oORF1'] = []
        states_idx['sense_start_oORF2'] = []
        states_idx['sense_sense_oORF1'] = []
        states_idx['sense_sense_oORF2'] = []
        states_idx['stop_sense_oORF1'] = []
        states_idx['stop_sense_oORF2'] = []
        states_idx['sense_stop_iORF1'] = []
        states_idx['sense_stop_iORF2'] = []
        states_idx['sense_iORF1'] = []
        states_idx['sense_iORF2'] = []

        altorfs_idx['start_oORF'] = []
        altorfs_idx['sense_oORF'] = []
        altorfs_idx['stop_oORF'] = []
        altorfs_idx['stop_iORF'] = []
        altorfs_idx['oORF']
        altorfs_idx['iORF']

        for codon1 in sorted(sense_codon_freqs):

            # Add +1 and +2 overlapping start codon states
            for codon2 in sorted(start_codon_freqs):
                if codon1[1] == codon2[0] and codon1[2] == codon2[1]:
                    states += [(codon1 + '_' + codon2 + '#')]
                    states_idx['sense_start_oORF1'] += [idx]
                    altorfs_idx['start_oORF'] += [idx]
                    altorfs_idx['oORF'] += [idx]
                    idx += 1
                if codon1[2] == codon2[0]:
                    states += [(codon1 + '_' + codon2 + '#')]
                    states_idx['sense_start_oORF2'] += [idx]
                    altorfs_idx['start_oORF'] += [idx]
                    altorfs_idx['oORF'] += [idx]
                    idx += 1

            # Add +1 and +2 overlapping sense codon states
            for codon2 in sorted(sense_codon_freqs):
                if codon1[1] == codon2[0] and codon1[2] == codon2[1]:
                    states += [(codon1 + '_' + codon2)]
                    states_idx['sense_sense_oORF1'] += [idx]
                    altorfs_idx['sense_oORF'] += [idx]
                    altorfs_idx['oORF'] += [idx]
                    idx += 1
                if codon1[2] == codon2[0]:
                    states += [(codon1 + '_' + codon2)]
                    states_idx['sense_sense_oORF2'] += [idx]
                    altorfs_idx['sense_oORF'] += [idx]
                    altorfs_idx['oORF'] += [idx]
                    idx += 1

            # Add +1 and +2 overlapping stop codon states
            for codon2 in sorted(stop_codon_freqs):
                if codon2[1] == codon1[0] and codon2[2] == codon1[1]:
                    states += [(codon2 + '*' + '_' + codon1)]
                    states_idx['stop_sense_oORF1'] += [idx]
                    altorfs_idx['stop_oORF'] += [idx]
                    altorfs_idx['oORF'] += [idx]
                    idx += 1
                if codon2[2] == codon1[0]:
                    states += [(codon2 + '*' + '_' + codon1)]
                    states_idx['stop_sense_oORF1'] += [idx]
                    altorfs_idx['stop_oORF'] += [idx]
                    altorfs_idx['oORF'] += [idx]
                    idx += 1

            # Add +1 and +2 internal stop codon states
            for codon2 in sorted(stop_codon_freqs):
                if codon1[1] == codon2[0] and codon1[2] == codon2[1]:
                    states += [(codon1 + '_' + codon2 + '*')]
                    states_idx['sense_stop_iORF1'] += [idx]
                    altorfs_idx['stop_iORF'] += [idx]
                    altorfs_idx['iORF'] += [idx]
                    idx += 1
                if codon1[2] == codon2[0]:
                    states += [(codon1 + '_' + codon2 + '*')]
                    states_idx['sense_stop_iORF1'] += [idx]
                    altorfs_idx['stop_iORF'] += [idx]
                    altorfs_idx['iORF'] += [idx]
                    idx += 1

            # Add +1 and +2 extra sense codon states after iORF
            states += [(codon1)]
            states_idx['sense_iORF1'] += [idx]
            idx += 1
            states += [(codon1)]
            states_idx['sense_iORF2'] += [idx]
            idx += 1

    return states, states_idx, altorfs_idx



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# GFF/GTF READER
# Author: Adapted from code by Kamil Slowikowski
# available from https://gist.github.com/slowkow/8101481
# Date: 2022.11.02
# -----------------------------------------------------------------------------

import re
import gzip
import logging
import pandas as pd
from collections import defaultdict

logger = logging.getLogger(__name__)


GTF_HEADER  = ['chrom', 'source', 'feature', 'start', 'end', 'score',
               'strand', 'frame']
R_SEMICOLON = re.compile(r'\s*;\s*')
R_COMMA     = re.compile(r'\s*,\s*')
R_KEYVALUE  = re.compile(r'(\s+|\s*=\s*)')


def gtf_to_df(file):
    """
    Open an optionally gzipped GTF file and return a pandas DataFrame
    GTF file format: ensembl.org/info/website/upload/gff.html
    Args:
        file: path to an annotations gff/gtf file
    Returns:
        pandas DataFrame with the annotations
    Raises:
        IOError: an error occurred accessing the annotations file
        IndexError, TypeError, KeyError: an error occurred reading the
            annotations file
    """
    # Each column is a list stored as a value in this dict
    annotations = defaultdict(list)

    for i, line in enumerate(lines(file)):
        for key in line.keys():
            # This key has not been seen yet, so set it to None for all
            # previous lines
            if key not in annotations:
                annotations[key] = [None] * i

        # Ensure this row has some value for each column
        for key in annotations.keys():
            annotations[key].append(line.get(key, None))

    return pd.DataFrame(annotations)


def lines(file):
    # Open the annotations gtf file for reading
    try:
        if file.endswith('.gz'): f = gzip.open(file,'r')
        else: f = open(file,'r')

    except IOError:
        logging.error('Could not open annotations file %s' % file)

    else:
        with f:
            try:
                for line in f:

                    # Skip header lines
                    if line.startswith('#'):
                        continue

                    # Parse annotation data lines
                    else:
                        yield parse(line)

            except (IndexError, TypeError, KeyError):
                logging.error(('Could not parse annotations file %s.') % file)
                raise SyntaxError


def parse(line):
    annotations = {}

    fields = line.rstrip().split('\t')

    for i, col in enumerate(GTF_HEADER):
        annotations[col] = _get_value(fields[i])

    # INFO field consists of "key1=value;key2=value;..."
    infos = [x for x in re.split(R_SEMICOLON, fields[8]) if x.strip()]

    for i, info in enumerate(infos, 1):
        # It should be key="value"
        try:
            key, _, value = re.split(R_KEYVALUE, info, 1)
        # But sometimes it is just "value"
        except ValueError:
            key = 'INFO{}'.format(i)
            value = info
        # Ignore the field if there is no value
        if value:
            annotations[key] = _get_value(value)

    return annotations


def _get_value(value):
    if not value:
        return None

    # Strip double and single quotes
    value = value.strip('"\'')

    # Return a list if the value has a comma
    if ',' in value:
        value = re.split(R_COMMA, value)

    # These values are equivalent to None
    elif value in ['', '.', 'NA']:
        return None

    return value



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# HMM plotting functions
# Author: Mary Richardson
# Date: 2021.01.05
# -----------------------------------------------------------------------------

import numpy as np
import scipy.stats as ss
import itertools as it
import matplotlib.pyplot as plt
from matplotlib.ticker import FormatStrFormatter as fmt

from orfeus.hmm.run import hmm_single, get_annotated_path
from orfeus.output.text import readable_sequence, trace_dump

# Customize plotting style
colors = ['#6F6F6F',
          '#DDDDDD',
          '#E0BA4D',
          '#9ABC61',
          '#E3995A',
          '#5385AA',
          '#C55A56',
          '#A472A5']
basics = ['#5385AA',
          '#C55A56',
          '#6F6F6F']


def calculate_metagene_signal(data_df, w):
    """
    Calculate the mean signal upstream/downstream of the start and stop codon
    Args:
        data_df: processed data
        w: window of nucleotides upstream/downstream of the start or stop codon
            to return
    Returns:
        signal_start: mean signal upstream/downstream of the first ORF nt
        signal_stop: mean signal upstream/downstream of the last ORF nt
    Raises:

    """
    # Default zeros
    signal_5UTR = np.zeros(w)
    signal_3UTR = np.zeros(w)

    # Group the data by transcript then feature
    df = data_df.groupby(['transcript_id', 'feature']) \
                .agg(list).reset_index()

    # Get the reads by feature
    data_ORF = df[df['feature']=='ORF']['norm_reads'].tolist()
    data_5UTR = df[df['feature']=='5UTR']['norm_reads'].tolist()
    data_3UTR = df[df['feature']=='3UTR']['norm_reads'].tolist()

    # Calculate the total number of transcripts
    tx_count = len(data_ORF)

    # START codon signal
    if len(data_5UTR) > 0:
        signal_5UTR = np.sum(np.array([([0]*w + r)[-w:] for r in data_5UTR]), \
                             axis=0)
    signal_ORF = np.sum(np.array([(r + [0]*w)[:w] for r in data_ORF]), \
                        axis=0)

    # Aggregate START signal
    signal_start = np.concatenate((signal_5UTR, signal_ORF))
    signal_start /= tx_count

    # STOP codon
    signal_ORF = np.sum(np.array([([0]*w + r)[-w:] for r in data_ORF]), \
                       axis=0)
    if len(data_3UTR) > 0:
        signal_3UTR = np.sum(np.array([(r + [0]*w)[:w] for r in data_3UTR]), \
                            axis=0)

    # Aggregate STOP signal
    signal_stop = np.concatenate((signal_ORF, signal_3UTR))
    signal_stop /= tx_count

    return signal_start, signal_stop


def plot_metagene_signal(signal_start,
                         signal_stop,
                         w,
                         file):
    """
    Plot the mean signal upstream/downstream of the start and stop codon
    Args:
        signal_start: mean signal upstream/downstream of the first ORF nt
        signal_stop: mean signal upstream/downstream of the last ORF nt
        w: window of nucleotides upstream/downstream of the start or stop codon
            to return
        file: output plot file path
    Returns:

    Raises:

    """
    # Set the style and axes
    fig, (ax1,ax2) = plt.subplots(1, 2,
                                  sharex=False,
                                  sharey=True,
                                  figsize=(4.5,2.75))
    ax1.yaxis.set_major_formatter(fmt('%.2f'))

    # Plot the start signal
    signal_start = signal_start[-(w+15):]
    ax1.bar(np.linspace(-14, w, w+15),
            signal_start,
            color=basics,
            width=0.9)
    ax1.set_ylabel('Average rho', fontsize=12)
    ax1.set_xlabel('Position relative \nto start (nt)', fontsize=12)

    # Plot the stop signal
    signal_stop = signal_stop[:(w+15)]
    ax2.bar(np.linspace(-w, 14, w+15),
            signal_stop,
            color=basics,
            width = 0.9)
    ax2.set_xlabel('Position relative \nto stop (nt)', fontsize=12)

    # Save and show the plot
    plt.tight_layout()
    plt.savefig(file+'.pdf', format='pdf')
    plt.savefig(file+'.eps', format='eps')
    plt.show()


def plot_emissions(data_df,
                   states,
                   file):
    """
    Plots a histogram of rho values for each type of canonical state
        (5'UTR, 3'UTR,
        start 1st nt, start 2nd nt, start 3rd nt,
        sense 1st nt, sense 2nd nt, sense 3rd nt,
        stop 1st nt, stop 2nd nt, stop 3rd nt)
    Args:
        data_df: processed data
        states: list of state names
        file: output plot file path
    Returns:

    Raises:

    """
    # Regroup the data by subcodon nucleotide position
    data_df['nt_pos'] = data_df['nt_pos'].astype(str)
    df = data_df.groupby(['codon_type', 'nt_pos']).agg(list).reset_index()
    df = df.sort_values('codon_type', ascending=False)

    # Set up the figure
    fig = plt.figure(figsize=(9,10))

    # Iterate through all rows of the emission distribution dataframe
    for i, (name, row) in enumerate(df.iterrows()):
        ax = plt.subplot(4, 3, i+1)

        # Get the reads data for this row (excluding values that are -inf)
        reads = np.array(row['norm_reads'])

        # Plot the reads data
        if i<9: color = basics[i%3]
        else: color = 'grey'
        ax.hist(reads, bins=np.linspace(xlim[0], xlim[1], 30),
                alpha=0.8, density=True, color=color)

        # Label the state
        ax.text(3.5, ylim[1]-ylim[1]*.2, row['codon_type'],
                fontsize=12, weight='bold', color=color)

        # Label the axes
        ax.set_xlabel('Rho', fontsize=12)
        ax.set_ylabel('Frequency', fontsize=12)

    # Save and show the plot
    plt.tight_layout()
    plt.savefig(file+'.pdf', format='pdf')
    plt.savefig(file+'.eps', format='eps')
    plt.show()


def get_prediction(path_h1,
                   path_h2,
                   altorfs_idx):
    """
    Get the positions of each altORF event from the viterbi path
    Args:
        path_h1: predicted state path
        path_h2: annotated state path
        altorfs_idx: dictionary mapping more general state type to
            state indices (0-indexed)
    Returns:
        annot_ORF: boolean array representing annotated ORF positions
        UTR5: boolean array representing 5'UTR positions
        UTR3: boolean array representing 3'UTR positions
        ORF: boolean array representing ORF positions
        uORF: boolean array representing uORF positions
        dORF: boolean array representing dORF positions
        pPRF: boolean array representing +1 PRF positions
        mPRF: boolean array representing -1 PRF positions
        SCR: boolean array representing SCR positions
    Raises:

    """
    if len(path_h2)==0:
        annot_ORF = np.zeros(len(path_h1))
    else:
        annot_ORF = np.array([s in altorfs_idx['ORF'] for s in path_h2])
    UTR5 = np.array([s==altorfs_idx['5UTR'] for s in path_h1])
    UTR3 = np.array([s==altorfs_idx['3UTR'] for s in path_h1])
    uORF = np.array([(s in altorfs_idx['uORF']) for s in path_h1])
    dORF = np.array([(s in altorfs_idx['dORF']) for s in path_h1])
    pPRF = np.array([(path_h1[i] in altorfs_idx['PRF']) and \
                     (path_h1[i+1] not in altorfs_idx['PRF']) \
            for i,s in enumerate(path_h1[:-1])] + [0])
    mPRF = np.array([(path_h1[i] in altorfs_idx['PRF']) and \
                     (path_h1[i+1] in altorfs_idx['PRF']) \
            for i,s in enumerate(path_h1[:-1])] + [0])
    SCR = np.array([(path_h1[i] in altorfs_idx['SCR']) and \
                    (path_h1[i+1] not in altorfs_idx['SCR']) \
            for i,s in enumerate(path_h1[:-1])] + [0])
    ORF = np.invert(np.logical_or(UTR5, UTR3))

    return annot_ORF, UTR5, UTR3, ORF, uORF, dORF, pPRF, mPRF, SCR


def plot_prediction(r,
                    path_h1,
                    path_h2,
                    altorfs_idx,
                    transcript_name,
                    file,
                    idx=0,
                    subcodon=True):
    """
    Plots the model prediction as shaded regions over the observed rho
        values for the given transcript
    Args:
        r: rho values
        path_h1: predicted state path
        path_h2: annotated state path
        altorfs_idx: dictionary mapping more general state type to
            state indices (0-indexed)
        transcript_name: name of the transcript
        file: output plot file path
        idx: true altORF position to plot for comparison, if there is one
        subcodon: whether to plot the signal broken out into subcodons 1, 2, 3
    Returns:

    Raises:

    """
    # Get the predicted positions
    annot_ORF, UTR5, UTR3, ORF, uORF, dORF, pPRF, mPRF, SCR = \
        get_prediction(path_h1,
                       path_h2,
                       altorfs_idx)
    start = sum(UTR5)-1

    if subcodon:

        # Split into subcodon signals
        nt1 = r.copy()
        nt1[start+1::3] = [0] * len(nt1[start+1::3])
        nt1[start+1::-3] = [0] * len(nt1[start+1::-3])
        nt1[start+2::3] = [0] * len(nt1[start+2::3])
        nt1[start+2::-3] = [0] * len(nt1[start+2::-3])

        nt2 = r.copy()
        nt2[start+0::3] = [0] * len(nt2[start+0::3])
        nt2[start+0::-3] = [0] * len(nt2[start+0::-3])
        nt2[start+2::3] = [0] *len(nt2[start+2::3])
        nt2[start+2::-3] = [0] * len(nt2[start+2::-3])

        nt3 = r.copy()
        nt3[start+0::3] = [0] * len(nt3[start+0::3])
        nt3[start+0::-3] = [0] * len(nt3[start+0::-3])
        nt3[start+1::3] = [0] * len(nt3[start+1::3])
        nt3[start+1::-3] = [0] * len(nt3[start+1::-3])

        # Create the plot
        fig,(ax1,ax2,ax3,ax4) = plt.subplots(4,1,
                                    gridspec_kw={'height_ratios': [1.5,1,1,1]},
                                    sharex=True,
                                    sharey=True,
                                    figsize=(5, 4))
        for ax in [ax2,ax3,ax4]:
            ax.spines['left'].set_visible(False)
            ax.spines['right'].set_visible(False)
            ax.spines['top'].set_visible(False)

            # Format the axes
            ax.set_xlim([0, len(r)])
            ax.set_ylim([0, max(r)+5])
            ax.yaxis.set_major_formatter(fmt('%.1f'))

        # Plot the signal
        ax1.plot(r, color='k', alpha=0.5)
        ax2.plot(nt1, color=colors[0], alpha=0.8)
        ax3.plot(nt2, color=colors[0], alpha=0.8)
        ax4.plot(nt3, color=colors[0], alpha=0.8)

        # Label the axes
        fig.add_subplot(111, frame_on=False)
        plt.tick_params(labelcolor='none', bottom=False, left=False)
        plt.xlabel('Position (nt)', fontsize=12)
        plt.ylabel('Rho', fontsize=12)

    else:

        # Create the plot
        fig,(ax1) = plt.subplots(1,1, figsize=(4, 2))

        # Plot the signal
        ax1.plot(r, color='k', alpha=0.5)

        # Format the axes
        ax1.set_xlim([0, len(r)])
        ax1.set_ylim([0, max(r)+5])
        ax1.yaxis.set_major_formatter(fmt('%.1f'))

        # Label the axes
        ax1.set_xlabel('Position (nt)', fontsize=12)
        ax1.set_ylabel('Rho', fontsize=12, rotation='vertical')

    # Title the figure
    plt.title(transcript_name, fontsize=14, weight='bold')

    # Save the unshaded plot
    plt.savefig(file+'.eps', format='eps')

    # Shade the annotated and predicted path on the first set of axes
    y = max(r)+5
    x = np.linspace(0, len(r), len(r), endpoint=False)
    ax1.fill_between(x, ORF*y,
                     facecolor=colors[1],
                     edgecolor=colors[1],
                     alpha=0.5,
                     linewidth=1)
    ax1.fill_between(x, uORF*y,
                     facecolor=colors[2],
                     edgecolor=colors[2],
                     alpha=0.5,
                     linewidth=1)
    ax1.fill_between(x, dORF*y,
                     facecolor=colors[2],
                     edgecolor=colors[2],
                     alpha=0.5,
                     linewidth=1)
    ax1.fill_between(x, pPRF*y,
                     facecolor = 'none',
                     edgecolor=colors[5],
                     linewidth=1)
    ax1.fill_between(x, mPRF*y,
                     facecolor = 'none',
                     edgecolor=colors[6],
                     linewidth=1)
    ax1.fill_between(x, SCR*y,
                     facecolor = 'none',
                     edgecolor=colors[7],
                     linewidth=1)
    ax1.fill_between(x, annot_ORF*y,
                     facecolor = 'none',
                     edgecolor=colors[1],
                     linewidth=1,
                     linestyle=':')

    # Plot the true location, if there is one
    if idx!=0:
        idx = start + idx
        ax1.axvline(idx, 0, y, color='k', linestyle=':')

    # Save and show the plot
    plt.savefig(file+'.pdf', format='pdf')
    plt.show()


def orfeus(data_df,
           transcript_id,
           transcript_name,
           parameters_h0,
           parameters_h1,
           idx=0,
           suppress=True,
           debug=False,
           subcodon=True):
    """
    Displays the predicted path and plots
    Args:
        data_df: transcript data
        transcript_id: ID of the transcript
        transcript_name: name of the transcript
        parameters_h1: list of all parameter matrices for the current model
        parameters_h0: list of all parameter matrices for the null model
        idx: true altORF position to plot for comparison, if there is one
        suppress: whether to suppress the path output
        debug: whether to print the full traceback
        subcodon: whether to plot the signal broken out into subcodons 1, 2, 3
    Returns:

    Raises:

    """
    # Unpack the parameters
    states, states_idx, altorfs_idx, \
    p_emit_nt_h1, p_emit_rho_h1, rho_ranges, p_trans_h1, p_trans_rev_h1, \
    prev_states_options_h1, next_states_options_h1 = parameters_h1

    # Extract just the transcript of interest
    df = data_df[data_df['transcript_id']==transcript_id]

    # Get the transcript coverage, sequence, and path
    c = df['mean_reads'].tolist()[0]
    x = df['seq'].tolist()
    r = df['log_reads'].tolist()
    path = get_annotated_path(df['state'].tolist(), states)

    print('Mean reads per nucleotide: %f' % c)

    path_h2, path_h1, path_h0, \
    path_score_h2, path_score_h1, path_score_h0, log_odds_score, \
    log_emit_h1, log_trans_h1 = hmm_single(x, r,
                                           path,
                                           parameters_h1,
                                           parameters_h0)

    # Print the annotated path, predicted path, and null path
    if not suppress:

        #seq_h2 = readable_sequence(x, path_h2, states_idx)
        #print('Annotated:               ', ''.join(seq_h2))

        #seq_h0 = readable_sequence(x, path_h0, states_idx)
        #print('Null:                    ', ''.join(seq_h0))

        seq_h1 = readable_sequence(x, path_h1, states_idx)
        print('Predicted:               ', ''.join(seq_h1))

        #print('Annotated path score:    %f' % path_score_h2)
        #print('Null path score:         %f' % path_score_h0)
        #print('Predicted path score:    %f' % path_score_h1)

    # Print the full trace for debugging
    if debug:
        trace_dump(x, r, path_h1, states, log_emit_h1, log_trans_h1)

    # Plot the prediction
    file = transcript_name
    plot_prediction(r,
                    path_h1,
                    path_h2,
                    altorfs_idx,
                    transcript_name,
                    file,
                    idx,
                    subcodon)



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# Set transition probabilities
# Author: Mary Richardson
# Date: 2020.07.07, last update 2023.03.02
# -----------------------------------------------------------------------------

import logging
import numpy as np

logger = logging.getLogger(__name__)

from set_parameters_helper_functions import get_codon, get_nt, normalize


def get_feature_freqs(data_df,
                      f_5UTR=None,
                      f_3UTR=None):
    """
    Calculate the fraction of ORFs with 5'UTRs and 3'UTRs
    Args:
        data_df: pandas DataFrame with all annotation info
        f_5UTR: (optional) float indicating the fraction of ORFs having 5'UTRs
            default is to infer from the input data
        f_3UTR: (optional) float indicating the fraction of ORFs having 3'UTRs
            default is to infer from the input data
    Returns:
        f_5UTR: float indicating the fraction of ORFs having 5'UTRs
        f_3UTR: float indicating the fraction of ORFs having 3'UTRs
    Raises:
        ValueError: No ORFs found in dataset
    """
    # Calculate the total number of genes
    counts_gene = data_df[data_df['feature']=='ORF']['transcript_id'].nunique()

    # Calculate the fraction of genes that start with 5UTRs
    if not f_5UTR:
        counts_5UTR = data_df[data_df['feature']=='5UTR']['transcript_id'].nunique()
        f_5UTR = counts_5UTR/counts_gene

    # Calculate the fraction of genes that end with 3UTRs
    if not f_3UTR:
        counts_3UTR = data_df[data_df['feature']=='3UTR']['transcript_id'].nunique()
        f_3UTR = counts_3UTR/counts_gene

    # Check that gene count is positive
    if counts_gene <= 0:
        f_5UTR  = 0
        f_3UTR = 0
        logging.error('No ORFs found in dataset. Check annotations!')

    return f_5UTR, f_3UTR


def adjust_length(len, orf=False):
    """
    Adjust the mean lengths of 5'UTRs, 3'UTRs, and ORFs
    Args:
        len: int mean length
        orf: boolean indicating whether this length corresponds to an ORF
    Returns:
        adjusted_len: int length converted from nt to codons if it an ORF
    Raises:
        ValueError: Invalid mean length for UTR or ORF
    """
    # Convert to codons and subtract the start and stop codons
    if orf:
        len = len/3 - 2

    # Don't allow zero or negative lengths
    if len <= 0:
        len = 1
        logging.warning(('Invalid mean length for UTR or ORF. '
            'Length set to 1 nucleotide for UTRs or 1 codon for ORFs.'))

    return len


def get_feature_lengths(data_df,
                        len_5UTR=None,
                        len_3UTR=None,
                        len_orf=None,
                        len_uorf=None,
                        len_dorf=None):
    """
    Calculate the mean lengths of 5'UTRs, 3'UTRs, and ORFs
    Args:
        data_df: pandas DataFrame with all annotation info
        len_5UTR: (optional) float indicating the mean length of 5'UTRs
            default is to infer from the input data
        len_3UTR: (optional) float indicating the mean length of 3'UTRs
            default is to infer from the input data
        len_orf: (optional) float indicating the mean length of ORFs
            default is to infer from the input data
        len_uorf: (optional) float indicating the mean length of uORFs
            default is to set equal to len_orf
        len_dorf: (optional) float indicating the mean length of dORFs
            default is to set equal to len_orf
    Returns:
        len_5UTR: float indicating the mean length of 5'UTRs
        len_3UTR: float indicating the mean length of 3'UTRs
        len_orf: loat indicating the mean length of ORFs
        len_uorf: float indicating the mean length of uORFs
        len_dorf: float indicating the mean length of dORFs
    """
    # Group by feature and gene so each feature is counted only once
    group = data_df.groupby(['feature','transcript_id']).mean().reset_index()

    # Calculate the mean length of each feature that is not given
    if not len_5UTR:
        len_5UTR = group[group['feature']=='5UTR']['feature_len'].mean()
    if not len_3UTR:
        len_3UTR = group[group['feature']=='3UTR']['feature_len'].mean()
    if not len_orf:
        len_orf = group[group['feature']=='ORF']['feature_len'].mean()
    if not len_uorf:
        len_uorf = len_uorf
    if not len_dorf:
        len_dorf = len_dorf

    # Aujust the length
    len_5UTR = adjust_length(len_5UTR)
    len_3UTR = adjust_length(len_3UTR)
    len_orf = adjust_length(len_orf, orf=True)
    len_uorf = adjust_length(len_uorf, orf=True)
    len_dorf = adjust_length(len_dorf, orf=True)

    return len_5UTR, len_3UTR, len_orf, len_uorf, len_dorf


def get_feature_codon_freqs(data_df,
                            feature,
                            feature_codons=None):
    """
    Calculate the frequency of each valid codon in this feature
    Args:
        data_df: pandas DataFrame with all annotation info
        feature: string representing the feature ('START','STOP', or 'ORF')
        feature_codons: (optional) list of valid codons for this feature,
            default is to infer from the input data
    Returns:
        codon_freqs: dict mapping each codon to its frequency in this feature
    Raises:
        ValueError: Codons not observed in annotations present in list of
            codons provided for this feature
    """
    # Get the codon frequencies for this feature
    if feature_codons:
        codon_freqs = {}
        for codon in feature_codons:
            # Get the codon frequencies with a pseudocount
            # (in case codons in the list are not observed in annotations)
            codon_freqs[codon] = len(data_df[(data_df['codon_type']==feature) &
                                    (data_df['codon_seq']==codon)]) + 1

            # Warn if unannotated codon included in list
            if codon_freqs[codon] == 1:
                logging.warning(('Codon not observed in annotations present '
                    'in list of %s codons provided: %s.') % (feature, codon))

        codon_freqs = normalize(codon_freqs)

    else:
        # Get the codon frequencies without a pseudocount
        # (only consider codons observed in annotations)
        codon_freqs = data_df[(data_df['codon_type']==feature)] \
                        ['codon_seq'].value_counts(normalize=True).to_dict()

    return codon_freqs


def get_codon_freqs(data_df,
                    codons,
                    start_codons=None,
                    sense_codons=None,
                    stop_codons=None):
    """
    Calculate the frequency of start, sense, and stop codons
    Args:
        data_df: pandas DataFrame with all annotation info
        codons: list of valid codons
        start_codons: (optional) list of valid start codons,
            default is to infer from the input data
        sense_codons: (optional) list of valid sense codons,
            default is to infer from the input data
        stop_codons: (optional) list of valid stop codons,
            default is to infer from the input data
    Returns:
        start_codon_freqs: dict mapping each start codon to its frequency
        sense_codon_freqs: dict mapping each sense codon to its frequency
        stop_codon_freqs: dict mapping each stop codon to its frequency
    Raises:
        ValueError: Invalid codons filtered out of dataset
    """
    # Exclude NaN codons
    data_df = data_df[data_df['codon_seq'].notnull()]
    unfiltered_len = len(data_df)

    # Only consider codons that are valid
    # (exclude 2nt or 1nt fragments and invalid nucleotides)
    data_df = data_df[data_df['codon_seq'].isin(codons)]
    filtered_len = len(data_df)

    # Check if invalid codons were filtered out
    if filtered_len < unfiltered_len:
        logging.warning('Invalid codons filtered out of dataset.')

    # Get codon frequencies for each feature (with a pseudocount)
    start_codon_freqs = \
        get_feature_codon_freqs(data_df, 'START', feature_codons=start_codons)
    sense_codon_freqs = \
        get_feature_codon_freqs(data_df, 'ORF', feature_codons=sense_codons)
    stop_codon_freqs = \
        get_feature_codon_freqs(data_df, 'STOP', feature_codons=stop_codons)

    return start_codon_freqs, sense_codon_freqs, stop_codon_freqs


def set_constants(data_df,
                  codons,
                  f_5UTR=None,
                  f_3UTR=None,
                  len_5UTR=None,
                  len_3UTR=None,
                  len_orf=None,
                  len_uorf=None,
                  len_dorf=None,
                  start_codons=None,
                  sense_codons=None,
                  stop_codons=None):
    """
    Set the constants based on the current data and annotations
    Args:
        data_df: pandas DataFrame of all annotated transcripts
        codons: list of valid codons
        f_5UTR: (optional) float indicating the fraction of ORFs having 5'UTRs
            default is to infer from the input data
        f_3UTR: (optional) float indicating the fraction of ORFs having 3'UTRs
            default is to infer from the input data
        len_5UTR: (optional) float indicating the mean length of 5'UTRs
            default is to infer from the input data
        len_3UTR: (optional) float indicating the mean length of 3'UTRs
            default is to infer from the input data
        len_orf: (optional) float indicating the mean length of ORFs
            default is to infer from the input data
        len_uorf: (optional) float indicating the mean length of uORFs
            default is to set equal to len_orf
        len_dorf: (optional) float indicating the mean length of dORFs
            default is to set equal to len_orf
        start_codons: (optional) list of valid start codons,
            default is to infer from the input data
        sense_codons: (optional) list of valid sense codons,
            default is to infer from the input data
        stop_codons: (optional) list of valid stop codons,
            default is to infer from the input data
    Returns:
        f_5UTR: float indicating the fraction of ORFs having 5'UTRs
        f_3UTR: float indicating the fraction of ORFs having 3'UTRs
        len_5UTR: float indicating the mean length of 5'UTRs
        len_3UTR: float indicating the mean length of 3'UTRs
        len_orf: float indicating the mean length of ORFs
        len_uorf: float indicating the mean length of uORFs
        len_dorf: float indicating the mean length of dORFs
        start_codon_freqs: dict mapping each start codon to its frequency
        sense_codon_freqs: dict mapping each sense codon to its frequency
        stop_codon_freqs: dict mapping each stop codon to its frequency
    Raises:

    """
    # Get frequency of 5'UTRs and 3'UTRs
    f_5UTR, f_3UTR = get_feature_freqs(data_df,
                                       f_5UTR,
                                       f_3UTR)

    # Get mean lengths of 5'UTRs, 3'UTRs, and ORFs
    len_5UTR, len_3UTR, len_orf, len_uorf, len_dorf = \
            get_feature_lengths(data_df,
                                len_5UTR,
                                len_3UTR,
                                len_orf,
                                len_uorf,
                                len_dorf)

    # Get start, sense, and stop codon frequencies
    start_codon_freqs, sense_codon_freqs, stop_codon_freqs = \
        get_codon_freqs(data_df,
                        codons,
                        start_codons,
                        sense_codons,
                        stop_codons)

    return f_5UTR, f_3UTR, len_5UTR, len_3UTR, len_orf, len_uorf, len_dorf, \
            start_codon_freqs, sense_codon_freqs, stop_codon_freqs


def set_transition_probs(states,
                     states_idx,
                     f_5UTR,
                     f_3UTR,
                     len_5UTR,
                     len_3UTR,
                     len_orf,
                     len_uorf,
                     len_dorf,
                     start_codon_freqs,
                     sense_codon_freqs,
                     stop_codon_freqs,
                     parameters):
    """
    Set the probability of transitioning from each state to each other state
    Args:
        states: list of all states in the model
        states_idx: dictionary mapping each general state type to its
             state index (1-indexed)
        f_5UTR: float indicating the fraction of ORFs having 5'UTRs
        f_3UTR: float indicating the fraction of ORFs having 3'UTRs
        len_5UTR: float indicating the mean length of 5'UTRs
        len_3UTR: float indicating the mean length of 3'UTRs
        len_orf: float indicating the mean length of ORFs
        len_uorf: float indicating the mean length of uORFs
        len_dorf: float indicating the mean length of dORFs
        start_codon_freqs: dict mapping each start codon to its frequency
        stop_codon_freqs: dict mapping each stop codon to its frequency
        sense_codon_freqs: dict mapping each sense codon to its frequency
        parameters: altORF probability parameters
    Returns:
        p_trans: the transition probability matrix
    Raises:
        ValueError: Invalid parameter values
        ValueError: Invalid transition probability matrix
    """
    # Unpack parameters
    [alpha, beta, delta, gamma, zeta] = parameters

    # Don't allow invalid parameters
    for parameter in parameters:
        if parameter < 0 or parameter >= 1:
            logging.error(('Invalid parameter provided. Check that alpha, '
                'beta, gamma, delta, and zeta are each greater than or equal '
                'to zero and less than one.'))

    # Prepare the MxM transition matrix (add one to account for begin state)
    M = len(states)
    p_trans = np.zeros((M+1,M+1))


    ################ UTR states

    # Prob of transitioning from begin to 5UTR
    p_trans[0][states_idx['5UTR']+1] = f_5UTR

    # Prob of transitioning from 5UTR to 5UTR
    p_trans[states_idx['5UTR']+1][states_idx['5UTR']+1] = \
        (len_5UTR-1)/len_5UTR * (1-delta)

    # Prob of transitioning from 3UTR to 3UTR
    p_trans[states_idx['3UTR']+1][states_idx['3UTR']+1] = \
        (len_3UTR-1)/len_3UTR * (1-delta) * (1-zeta)

    # Prob of transitioning from 3UTR to end
    p_trans[states_idx['3UTR']+1][0] = 1/len_3UTR * (1-delta)


    ################ Normal states

    # For each START codon row of the transition matrix, update the values
    for i in states_idx['start']:
        codon = get_codon(i-1, states) # Sequence of this codon

        # Prob of transitioning from begin
        p_trans[0][i] = (1-f_5UTR) * start_codon_freqs[codon]

        # Prob of transitioning from 5UTR
        p_trans[states_idx['5UTR']+1][i] = \
            (1/len_5UTR) * start_codon_freqs[codon] * (1-delta)

        # Prob of transitioning from 3UTR(multiple ORFs)
        p_trans[states_idx['3UTR']+1][i] = \
            (len_3UTR-1)/len_3UTR * start_codon_freqs[codon] * (1-delta) * zeta

        # Prob of transitioning to the next SENSE codon
        for j in states_idx['sense']:
            next_codon = get_codon(j-1, states) # Sequence of this codon
            p_trans[i][j] = sense_codon_freqs[next_codon]

    # For each SENSE codon row of the transition matrix, update the values
    for i in states_idx['sense']:
        codon = get_codon(i-1, states) # Sequence of this codon

        # Prob of transitioning to the next SENSE codon
        for j in states_idx['sense']:
            next_codon = get_codon(j-1, states) # Sequence of this codon
            p_trans[i][j] = \
                (len_orf-1)/len_orf * sense_codon_freqs[next_codon] * (1-alpha)

        # Prob of transitioning to the next STOP codon
        for j in states_idx['stop']:
            next_codon = get_codon(j-1, states) # Sequence of this codon
            p_trans[i][j] = \
                1/len_orf * stop_codon_freqs[next_codon] * (1-gamma)

        # Prob of transitioning to the next SCR codon
        if gamma > 0:
            for j in states_idx['stop_SCR']:
                next_codon = get_codon(j-1, states) # Sequence of this codon
                p_trans[i][j] = \
                    1/len_orf * stop_codon_freqs[next_codon] * gamma

        # Prob of transitioning to X1
        if alpha > 0:
            p_trans[i][states_idx['X1']+1] = \
                (len_orf-1)/len_orf * alpha

        # Prob of transitioning from X1
        if alpha > 0:
            p_trans[states_idx['X1']+1][i] = \
                sense_codon_freqs[codon] * (1-beta)

        # Prob of transitioning from X2
        if beta > 0:
            p_trans[states_idx['X2']+1][i] = sense_codon_freqs[codon]

    # For each STOP codon row of the transition matrix, update the values
    for i in states_idx['stop']:
        codon = get_codon(i-1, states) # Sequence of this codon

        # Prob of transitioning to 3UTR
        p_trans[i][states_idx['3UTR']+1] = f_3UTR

        # Prob of transitioning to end
        p_trans[i][0] = (1-f_3UTR)


    ################ SCR states
    # For each SCR codon row of the transition matrix, update the values
    if gamma > 0:
        for i in states_idx['stop_SCR']:
            codon = get_codon(i-1, states) # Sequence of this codon

            # Prob of transitioning to the next SENSE codon
            for j in states_idx['sense']:
                next_codon = get_codon(j-1, states) # Sequence of this codon
                p_trans[i][j] = sense_codon_freqs[next_codon]


    ################ PRF states
    if alpha > 0 and beta > 0:
        # Prob of transitioning from X1 to X2
        p_trans[states_idx['X1']+1][states_idx['X2']+1] = beta


    ################ uORF states
    if delta > 0:
        # For each START codon row of the transition matrix, update the values
        for i in states_idx['start_uORF']:
            codon = get_codon(i-1, states) # Sequence of this codon

            # Prob of transitioning from 5UTR
            p_trans[states_idx['5UTR']+1][i] = \
                start_codon_freqs[codon] * delta

            # Prob of transitioning to the next SENSE codon
            for j in states_idx['sense_uORF']:
                next_codon = get_codon(j-1, states) # Sequence of this codon
                p_trans[i][j] = sense_codon_freqs[next_codon]

        # For each SENSE codon row of the transition matrix, update the values
        for i in states_idx['sense_uORF']:
            codon = get_codon(i-1, states) # Sequence of this codon

            # Prob of transitioning to the next SENSE codon
            for j in states_idx['sense_uORF']:
                next_codon = get_codon(j-1, states) # Sequence of this codon
                p_trans[i][j] = \
                    (len_uorf-1)/len_uorf * sense_codon_freqs[next_codon]

            # Prob of transitioning to the next STOP codon
            for j in states_idx['stop_uORF']:
                next_codon = get_codon(j-1, states) # Sequence of this codon
                p_trans[i][j] = \
                    1/len_uorf * stop_codon_freqs[next_codon]

        # For each STOP codon row of the transition matrix, update the values
        for i in states_idx['stop_uORF']:
            codon = get_codon(i-1, states) # Sequence of this codon

            # Prob of transitioning to 5UTR
            p_trans[i][states_idx['5UTR']+1] = 1


    ################ dORF states
    if delta > 0:
        # For each START codon row of the transition matrix, update the values
        for i in states_idx['start_dORF']:
            codon = get_codon(i-1, states) # Sequence of this codon

            # Prob of transitioning from 3UTR
            p_trans[states_idx['3UTR']+1][i] = \
                start_codon_freqs[codon] * delta

            # Prob of transitioning to the next SENSE codon
            for j in states_idx['sense_dORF']:
                next_codon = get_codon(j-1, states) # Sequence of this codon
                p_trans[i][j] = sense_codon_freqs[next_codon]

        # For each SENSE codon row of the transition matrix, update the values
        for i in states_idx['sense_dORF']:
            codon = get_codon(i-1, states) # Sequence of this codon

            # Prob of transitioning to the next SENSE codon
            for j in states_idx['sense_dORF']:
                next_codon = get_codon(j-1, states) # Sequence of this codon
                p_trans[i][j] = \
                    (len_dorf-1)/len_dorf * sense_codon_freqs[next_codon]

            # Prob of transitioning to the next STOP codon
            for j in states_idx['stop_dORF']:
                next_codon = get_codon(j-1, states) # Sequence of this codon
                p_trans[i][j] = 1/len_dorf * stop_codon_freqs[next_codon]

        # For each STOP codon row of the transition matrix, update the values
        for i in states_idx['stop_dORF']:
            codon = get_codon(i-1, states) # Sequence of this codon

            # Prob of transitioning to 3UTR
            p_trans[i][states_idx['3UTR']+1] = 1


    ################ oORF states
    if epsilon > 0:
        # For each +1 OVERLAP START codon row of the transition matrix, update the values
        for i in states_idx['sense_start_oORF1']:
            codon1, codon2 = get_overlap_codons(i-1, states)

            # Prob of transitioning from sense codon
            for j in states_idx['sense']:
                p_trans[i][j] = epsilon * \
                    sense_codon_freqs[codon1] * start_codon_freqs[codon2]

            # Prob of transitioning to next +1 oORF sense codon
            for j in states_idx['sense_sense_oORF1']:
                codon3, codon4 = get_overlap_codons(j-1, states)
                if codon2[2]==codon3[0]:
                    p_trans[i][j] = \
                        sense_codon_freqs[codon3] * sense_codon_freqs[codon4]

            p_trans[i][i+1] =

            # Prob of transitioning from overlapped nt2/nt1 to overlapped nt3/nt2 within this codon
            p_trans[i+1][i+2] = 1


        for i in range(states_idx['oORF1_sense_start'][0]+1, states_idx['oORF1_sense_start'][-1]+1, 2):
            sense_codon, start_codon = get_overlap_codons(i-1, states) # Sequences of these codons

            # Prob of transitioning from nt1 to nt2 within this codon
            p_trans[i][i+1] = 1

            # Prob of transitioning from nt2 to nt3 within this codon
            for j in range(states_idx['oORF1_start3'][0]+1, states_idx['oORF1_start3'][1]+1):
                next_sense_codon, next_start_codon = get_overlap_codons(j-1, states) # Sequences of these codons
                if start_codon==next_start_codon:
                    p_trans[i+1][j] = overlap1_start_sense_freqs[start_codon][next_sense_codon]

        # For each +2 OVERLAP START codon row of the transition matrix, update the values
        for i in range(states_idx['oORF2_start_extra'][0]+1, states_idx['oORF2_start_extra'][-1]+1, 2):
            sense_codon = get_codons(i-1, states) # Sequences of these codons


        for i in range(states_idx['oORF2_start23'][0]+1, states_idx['oORF2_start23'][1]+1, 2):
            sense_codon, start_codon = get_overlap_codons(i-1, states) # Sequences of these codons

            # Prob of transitioning from nt2 to nt3 within this codon
            p_trans[i][i+1] = 1

            # Prob of transitioning from nt1 to nt2 within this codon
            for j in range(states_idx['oORF2_start1'][0]+1, states_idx['oORF2_start1'][1]+1):
                prev_sense_codon, prev_start_codon = get_overlap_codons(j-1, states) # Sequences of these codons
                if start_codon==prev_start_codon:
                    p_trans[j][i] = overlap2_start_sense_freqs[start_codon][sense_codon]

            # Prob of transitioning into nt3 of this codon to OVERLAP SENSE
            for j in range(states_idx['oORF2_sense1'][0]+1, states_idx['oORF2_sense1'][1]+1):
                next_sense_codon1, next_sense_codon2 = get_overlap_codons(j-1, states) # Sequences of these codons
                if sense_codon==next_sense_codon1:
                    p_trans[i+1][j] = overlap2_sense1_freqs[sense_codon][next_sense_codon2]

        # For each +1 OVERLAP SENSE codon row of the transition matrix, update the values
        for i in range(states_idx['oORF1_sense12'][0]+1, states_idx['oORF1_sense12'][1]+1, 2):
            sense_codon1, sense_codon2 = get_overlap_codons(i-1, states) # Sequences of these codons

            # Prob of transitioning into nt1 of this codon from OVERLAP START
            for j in range(states_idx['oORF1_start3'][0]+1, states_idx['oORF1_start3'][1]+1):
                prev_sense_codon, prev_start_codon = get_overlap_codons(j-1, states) # Sequences of these codons
                if sense_codon1==prev_sense_codon:
                    p_trans[j][i] = overlap1_sense12_freqs[sense_codon1][sense_codon2]

            # Prob of transitioning into nt1 of this codon from OVERLAP SENSE
            for j in range(states_idx['oORF1_sense3'][0]+1, states_idx['oORF1_sense3'][1]+1):
                prev_sense_codon1, prev_sense_codon2 = get_overlap_codons(j-1, states) # Sequences of these codons
                if sense_codon1==prev_sense_codon1:
                    if sense_codon1 in overlap1_sense_stop_freqs:
                        p_trans[j][i] = (len_orf-1)/len_orf * overlap1_sense12_freqs[sense_codon1][sense_codon2]
                    else:
                        p_trans[j][i] = overlap1_sense12_freqs[sense_codon1][sense_codon2]

            # Prob of transitioning from nt1 to nt2 within this codon
            p_trans[i][i+1] = 1

            # Prob of transitioning from nt2 to the next OVERLAP SENSE codon
            for j in range(states_idx['oORF1_sense3'][0]+1, states_idx['oORF1_sense3'][1]+1):
                next_sense_codon1, next_sense_codon2 = get_overlap_codons(j-1, states) # Sequences of these codons
                if sense_codon2==next_sense_codon2:
                    if sense_codon2 in overlap2_sense_stop_freqs:
                        p_trans[i+1][j] = (len_orf-1)/len_orf * overlap1_sense3_freqs[sense_codon2][next_sense_codon1]
                    else:
                        p_trans[i+1][j] = overlap1_sense3_freqs[sense_codon2][next_sense_codon1]

        # For each +2 OVERLAP SENSE codon row of the transition matrix, update the values
        for i in range(states_idx['oORF2_sense23'][0]+1, states_idx['oORF2_sense23'][1]+1, 2):
            sense_codon1, sense_codon2 = get_overlap_codons(i-1, states) # Sequence of these codons

            # Prob of transitioning into nt2 of this codon from OVERLAP SENSE
            for j in range(states_idx['oORF2_sense1'][0]+1, states_idx['oORF2_sense1'][1]+1):
                prev_sense_codon1, prev_sense_codon2 = get_overlap_codons(j-1, states) # Sequences of these codons
                if sense_codon2==prev_sense_codon2:
                    if sense_codon2 in overlap1_sense_stop_freqs:
                        p_trans[j][i] = (len_orf-1)/len_orf * overlap2_sense23_freqs[prev_sense_codon2][sense_codon1]
                    else:
                        p_trans[j][i] = overlap2_sense23_freqs[prev_sense_codon2][sense_codon1]

            # Prob of transitioning from nt2 to nt3 within this codon
            p_trans[i][i+1] = 1

            # Prob of transitioning from nt3 to the next OVERLAP SENSE codon
            for j in range(states_idx['oORF2_sense1'][0]+1, states_idx['oORF2_sense1'][1]+1):
                next_sense_codon1, next_sense_codon2 = get_overlap_codons(j-1, states) # Sequences of these codons
                if sense_codon1==next_sense_codon1:
                    if sense_codon1 in overlap2_sense_stop_freqs:
                        p_trans[i+1][j] = (len_orf-1)/len_orf * overlap2_sense1_freqs[sense_codon1][next_sense_codon2]
                    else:
                        p_trans[i+1][j] = overlap2_sense1_freqs[sense_codon1][next_sense_codon2]

    if epsilon > 0:
        # For each +1 OVERLAP STOP codon row of the transition matrix, update the values
        for i in range(states_idx['oORF2_stop1'][0]+1, states_idx['oORF2_stop1'][1]+1):
            sense_codon, stop_codon = get_overlap_codons(i-1, states) # Sequences of these codons

            # Prob of transitioning into nt1 of this codon from OVERLAP SENSE
            for j in range(states_idx['oORF1_sense12'][0]+1, states_idx['oORF1_sense12'][1]+1, 2):
                prev_sense_codon1, prev_sense_codon2 = get_overlap_codons(j-1, states) # Sequences of these codons
                if sense_codon==prev_sense_codon2:
                    if sense_codon in overlap2_sense_stop_freqs:
                        p_trans[j+1][i] = 1/len_orf * overlap2_sense_stop_freqs[sense_codon][stop_codon]

        for i in range(states_idx['oORF2_stop23'][0]+1, states_idx['oORF2_stop23'][1]+1, 2):
            sense_codon, stop_codon = get_overlap_codons(i-1, states) # Sequences of these codons

            # Prob of transitioning into nt1 of this codon from OVERLAP STOP
            for j in range(states_idx['oORF2_stop1'][0]+1, states_idx['oORF2_stop1'][1]+1):
                prev_sense_codon, prev_stop_codon = get_overlap_codons(j-1, states) # Sequences of these codons
                if stop_codon==prev_stop_codon:
                    p_trans[j][i] = overlap2_stop_freqs[stop_codon][sense_codon]

            # Prob of transitioning from nt2 to nt3 within this codon
            p_trans[i][i+1] = 1

            # Prob of transitioning from nt3 to nt3 in the next SENSE codon
            for j in range(states_idx['sense'][0]+1, states_idx['sense'][1]+1, 3):
                next_codon = get_codon(j-1, states) # Sequences of these codons
                if sense_codon==next_codon:
                    p_trans[i+1][j+2] = 1

        # For each +2 OVERLAP STOP codon row of the transition matrix, update the values
        for i in range(states_idx['oORF1_stop12'][0]+1, states_idx['oORF1_stop12'][1]+1, 2):
            sense_codon, stop_codon = get_overlap_codons(i-1, states) # Sequences of these codons

            # Prob of transitioning into nt1 of this codon from OVERLAP SENSE
            for j in range(states_idx['oORF2_sense1'][0]+1, states_idx['oORF2_sense1'][1]+1):
                prev_sense_codon1, prev_sense_codon2 = get_overlap_codons(j-1, states) # Sequences of these codons
                if sense_codon==prev_sense_codon2:
                    if sense_codon in overlap1_sense_stop_freqs:
                        p_trans[j][i] = 1/len_orf * overlap1_sense_stop_freqs[sense_codon][stop_codon]

            # Prob of transitioning from nt1 to nt2 within this codon
            p_trans[i][i+1] = 1

        for i in range(states_idx['oORF1_stop3'][0]+1, states_idx['oORF1_stop3'][1]+1):
            sense_codon, stop_codon = get_overlap_codons(i-1, states) # Sequences of these codons

            # Prob of transitioning into nt1 of this codon from OVERLAP STOP
            for j in range(states_idx['oORF1_stop12'][0]+1, states_idx['oORF1_stop12'][1]+1, 2):
                prev_sense_codon, prev_stop_codon = get_overlap_codons(j-1, states) # Sequences of these codons
                if stop_codon==prev_stop_codon:
                    p_trans[j+1][i] = overlap1_stop_freqs[stop_codon][sense_codon]

            # Prob of transitioning from nt3 to nt2 in the next SENSE codon
            for j in range(states_idx['sense'][0]+1, states_idx['sense'][1]+1, 3):
                next_codon = get_codon(j-1, states) # Sequences of these codons
                if sense_codon==next_codon:
                    p_trans[i][j+1] = 1

    ################ iORF states
    if epsilon > 0:
        # For each +1 INTERNAL OVERLAP STOP codon row of the transition matrix, update the values
        for i in states_idx['iORF1_stop12']:
            sense_codon, stop_codon = get_overlap_codons(i-1, states) # Sequences of these codons

            # Prob of transitioning into nt1 of this codon from OVERLAP SENSE
            for j in range(states_idx['oORF1_sense3'][0]+1, states_idx['oORF1_sense3'][1]+1):
                prev_sense_codon1, prev_sense_codon2 = get_overlap_codons(j-1, states) # Sequences of these codons
                if sense_codon==prev_sense_codon1:
                    if sense_codon in overlap1_sense_stop_freqs:
                        p_trans[j][i] = 1/len_orf * overlap1_sense_stop_freqs[sense_codon][stop_codon]

            # Prob of transitioning from nt1 to nt2 within this codon
            p_trans[i][i+1] = 1

        for i in range(states_idx['iORF1_stop3'][0]+1, states_idx['iORF1_stop3'][1]+1):
            sense_codon, stop_codon = get_overlap_codons(i-1, states) # Sequences of these codons

            # Prob of transitioning into nt1 of this codon from INTERNAL OVERLAP STOP
            for j in range(states_idx['iORF1_stop12'][0]+1, states_idx['iORF1_stop12'][1]+1, 2):
                prev_sense_codon, prev_stop_codon = get_overlap_codons(j-1, states) # Sequences of these codons
                if stop_codon==prev_stop_codon:
                    p_trans[j+1][i] = overlap1_stop_freqs[stop_codon][sense_codon]

            # Prob of transitioning from nt3 to nt2 in the next SENSE codon
            for j in range(states_idx['sense'][0]+1, states_idx['sense'][1]+1, 3):
                next_codon = get_codon(j-1, states) # Sequences of these codons
                if sense_codon==next_codon:
                    p_trans[i][j+1] = 1

        # For each +2 INTERNAL OVERLAP STOP codon row of the transition matrix, update the values
        for i in range(states_idx['iORF2_stop1'][0]+1, states_idx['iORF2_stop1'][1]+1):
            sense_codon, stop_codon = get_overlap_codons(i-1, states) # Sequences of these codons

            # Prob of transitioning into nt1 of this codon from OVERLAP SENSE
            for j in range(states_idx['oORF2_sense23'][0]+1, states_idx['oORF2_sense23'][1]+1, 2):
                prev_sense_codon1, prev_sense_codon2 = get_overlap_codons(j-1, states) # Sequences of these codons
                if sense_codon==prev_sense_codon1:
                    if sense_codon in overlap2_sense_stop_freqs:
                        p_trans[j+1][i] = 1/len_orf * overlap2_sense_stop_freqs[sense_codon][stop_codon]

        for i in range(states_idx['iORF2_stop23'][0]+1, states_idx['iORF2_stop23'][1]+1, 2):
            sense_codon, stop_codon = get_overlap_codons(i-1, states) # Sequences of these codons

            # Prob of transitioning into nt1 of this codon from INTERNAL OVERLAP STOP
            for j in range(states_idx['iORF2_stop1'][0]+1, states_idx['iORF2_stop1'][1]+1):
                prev_sense_codon, prev_stop_codon = get_overlap_codons(j-1, states) # Sequences of these codons
                if stop_codon==prev_stop_codon:
                    p_trans[j][i] = overlap2_stop_freqs[stop_codon][sense_codon]

            # Prob of transitioning from nt2 to nt3 within this codon
            p_trans[i][i+1] = 1

            # Prob of transitioning from nt3 to nt3 in the next SENSE codon
            for j in range(states_idx['sense'][0]+1, states_idx['sense'][1]+1, 3):
                next_codon = get_codon(j-1, states) # Sequences of these codons
                if sense_codon==next_codon:
                    p_trans[i+1][j+2] = 1


    # Check that transition probabilities leaving each state sum to 1
    if not np.allclose(np.sum(p_trans, axis=1), np.ones(len(p_trans))):
        logging.error(('Invalid transition probability matrix. '
                    'Transitions leaving each state do not sum to 1.'))

    return p_trans


def rev_transition_probs(p_trans):
    """
    Calculate the the reverse transition probabilities for generating a
    sequence backwards
    Args:
        p_trans: the transition probability matrix
    Returns:
        p_trans_rev: the reverse transition probability matrix
    Raises:
        ValueError: Could not generate reverse transition matrix
    """
    # Calculate the stationary distribution (pi)
    try:
        # Calculate the eigenvector for the transpose of the transition probs
        w, v = np.linalg.eig(p_trans.T)

        # Find the eigenvalue that is approximately 1
        # calculate the difference array
        w1 = np.absolute(w-1.0).argmin()

        # Get the eigenvector corresponding to an eigenvalues of 1
        pi = np.real(v)[:,w1]

        # Normalize the eigenvector
        pi = pi/pi.sum()

        # Calcualte the backward transition probs
        M = len(p_trans)
        p_trans_rev = np.zeros([M,M])
        for j in range(M):
            for k in range(M):
                p_trans_rev[j,k] = pi[k] / pi[j] * p_trans[k,j]
        p_trans_rev = p_trans_rev/p_trans_rev.sum(axis=1, keepdims=True)

    except:
        logging.warning('Could not generate reverse transition matrix.')

    return p_trans_rev


def get_prev_states(p_trans):
    """
    Generate a list of all possible previous states for each state
    Args:
        p_trans: the transition probability matrix
    Returns:
        prev_state_options: list of indices of all possible
            (nonzero probability) previous states
    """
    prev_states_options = []

    for col in np.transpose(p_trans):
        this_prev_states = np.flatnonzero(col) # 1-indexed
        prev_states_options.append(this_prev_states)

    return np.array(prev_states_options, dtype=object)


def get_next_states(p_trans):
    """
    Generate a list of all possible next states for each state
    Args:
        p_trans: the transition probability matrix
    Returns:
        next_state_options: list of indices of all possible
            (nonzero probability) next states
    """
    next_states_options = []

    for row in p_trans:
        this_next_states = np.flatnonzero(row) # 1-indexed
        next_states_options.append(this_next_states)

    return np.array(next_states_options, dtype=object)



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# FASTA READER/WRITER
# Author: Mary Richardson
# Date: 2022.05.25
# -----------------------------------------------------------------------------

import gzip
import logging
import pandas as pd

logger = logging.getLogger(__name__)


def reverse_complement(seq):
    """
    Generates the reverse complement of a nucleotide sequence
    Args:
        seq: string representing a nucleotide sequence
    Returns:
        rev_comp: string representing the reverse complement nucleotide
            sequence
    Raises:
        KeyError: an error occurred finding the reverse complement
    """
    # Define the complement of each nucleotide
    complement = {'A':'T', 'C':'G', 'G':'C', 'T':'A', 'N':'N', 'R':'Y', 'Y':'R'}

    # Complement the nucleotide sequence
    try:
        comp = [complement[nt] for nt in seq]
    except KeyError:
        logging.error(('Invalid nucleotide in transcript sequence, '
            'cannot find reverse complement for sequence %s') % seq)

    # Reverse the nucleotide sequence
    rev_comp = comp[::-1]

    return ''.join(rev_comp)


def read_fasta(file, rev_comp=False):
    """
    Fetches sequence(s) from a fasta file
    Args:
        file: path to a fasta file
        rev_comp: (optional) whether to calculate the reverse complement
    Returns:
        seqs: dictionary mapping each sequence ID to its sequence
    Raises:
        IOError: an error occurred accessing the file
        IndexError, TypeError, KeyError: an error occurred reading the file
    """
    # Open the fasta file for reading
    try:
        if file.endswith('.gz'): f = gzip.open(file,'r')
        else: f = open(file,'r')

    except IOError:
        logging.error('Could not open fasta file %s for reading' % file)

    else:

        # Prepare a dictionary to store the seqs
        seqs = {}

        with f:
            for line in f:

                try:
                    # If this is a seq id line, start a new sequence
                    if line[0]=='>':
                        id = line.strip()[1:].split()[0]
                        seqs[id] = ''

                    # Otherwise, append to the current sequence
                    else:

                        # For negative strand, sequence is reverse complement
                        if rev_comp:
                            seqs[id] = reverse_complement(line.strip()) + seqs[id]

                        # For positive strand, sequence is unchanged
                        else:
                            seqs[id] = seqs[id] + line.strip()

                except:
                    logging.error('File %s is not in expected fasta format' \
                          % file)

    return seqs


def write_fasta(seqs: dict, file: str, chunk=60):
    """
    Writes a fasta file of sequences
    Args:
        seqs: dictionary mapping sequence IDs to sequences
        file: path to output fasta file
        chunk: max line length for sequence entries, add line breaks to
            maintain this max length
    Returns:

    Raises:
        IOError: an error occurred accessing the file
    """
    # Open the fasta file for writing
    try:
        f = open(file,'w')

    except IOError:
        logging.error('Could not open fasta file %s for writing' % file)

    else:

        with f:
            # Write all sequences
            for id in seqs:

                # Write the seq id
                f.write('>%s\n' % id)

                # Write the sequence with a fixed max line length
                i = 0
                while i < len(seqs[id])//chunk:
                    f.write('%s\n' % seqs[id][i*chunk:(i+1)*chunk])
                    i += 1
                f.write('%s\n' % seqs[id][i*chunk:])


def fasta_to_df(file, rev_comp=False):
    """
    Generate a DataFrame containing the nucleotide sequence at each
    position of each transcript
    Args:
        file: path to a fasta file containing a nucleotide sequence
            for each transcript
        rev_comp: (optional) whether to calculate the reverse complement
    Returns:
        pandas DataFrame with the transcript nucleotide sequences
    Raises:
        IOError: an error occurred accessing the transcriptome file
        IndexError, TypeError, KeyError: an error occurred reading the
            transcriptome file
    """
    # Create a dict of transcript sequences
    seqs = read_fasta(file, rev_comp)

    # Convert to a dataframe
    df = pd.DataFrame(seqs.items(), columns=['seq_id','seq'])
    df = df.assign(seq = df['seq'].str.split('')).explode('seq')
    df = df[df['seq'].astype(bool)]
    df['seq_pos'] = df.groupby('seq_id').cumcount()+1

    return df



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# Print ORFeus logo
# Author: Mary Richardson
# Date: 2023.03.31
# -----------------------------------------------------------------------------


CYAN = '\033[96m'
GREY = '\033[97m'
END = '\033[0m'

def print_logo():

    print('\n',
          '                                                                      ', CYAN+'//\\\\'+END,'\n',
          '                                                                     ', CYAN+'\\\\//'+END,'\n',
          '                                                                  ', CYAN+'//\\\\'+END,'\n',
          '                                                                  ', CYAN+'\\\\//'+END,'\n',
          '                                                                     ', CYAN+'//\\\\'+END,'\n',
          '                                                                     ', CYAN+'\\\\//'+END,'\n',
          '                                                                        ', CYAN+'//\\\\'+END,'\n',
          '                                                                        ', CYAN+'\\\\//'+END,'\n',
          '                                                                      ', CYAN+'//\\\\'+END,'\n',
          '                                                                      ', CYAN+'\\\\//'+END,'\n',
          '                                                                         ', CYAN+'//\\\\'+END,'\n',
          CYAN+'    ACGACG    GUCGUGC   CAGUCGGCA'+END, '                                       ', CYAN+'\\\\//'+END,'\n',
          CYAN+'  AGUCAGUCGU  ACGUUACG  UGACGUAAC'+END, '                                   ', GREY+'.RIBOSOMERIBOSOME.'+END,'\n',
          CYAN+' GUC      AUG CA    GUA AGC'+END, '        CGUAG   CAG   CGA  CAGUGC     ', GREY+'.RIBOSOMERIBOSOMERIBOSO.'+END,'\n',
          CYAN+' CA        GC UGAGUUGC  GUACUGAA'+END, ' GUA   AGC ACG   GAC CGU       ', GREY+'.MERIBOSOMERIBOSOMERIBOSOME.'+END,'\n',
          CYAN+' ACG      CGU GG ACGU   CGAGUAG'+END, '  ACGUAGUAG UCA   AUG  AGCA    ', GREY+'.RIBOSOMERIBOSOMERIBOSOMERIBO.'+END,'\n',
          CYAN+'  CGCGGGCAUG  AC   ACG  UGA'+END, '      GGA       CAU   GCU     GCA  ', GREY+'.SOMERIBOSOMERIBOSOMERIBOSOME.'+END,'\n',
          CYAN+'    GUCAGC    GC    GCC CAG'+END, '        AGUAG    CAGUGCA  CAUGCA     ', GREY+'.RIBOSOMERIBOSOMERIBOSOME.'+END,'\n',
          '                                                                  ', GREY+'.OMERIBOSOMERIBOSOMERIB.'+END,'\n',
          '##########################################################################################################','\n',
          '                                                                  ', GREY+'.RIBOSOMERIBOSOMERIBOSO.   '+END,'\n',
          '                                                                 ', GREY+'.MERIBOSOMERIBOSOMERIBOSO.  '+END,'\n',
          '                                                                    ', GREY+'.MERIBOSOMERIBOSOME.     '+END)



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# Build ORFeus model
# Author: Mary Richardson
# Date: 2023.03.03
# -----------------------------------------------------------------------------

import argparse
import logging
import os
import numpy as np

from orfeus.parsers.data import process_data, read_data_file
from orfeus.parameters.set import set_parameters
from orfeus.coverage.threshold import coverage_threshold_simulation
from orfeus.output.logo import print_logo


def main():

# ------------------------------------------------------------------------------
# Accept command line arguments
# ------------------------------------------------------------------------------

    parser = argparse.ArgumentParser(description='Build ORFeus model')

    # Input data
    parser.add_argument('plus_file', type=str,
                        help='riboseq plus strand bg/wig file path')
    parser.add_argument('minus_file', type=str,
                        help='riboseq minus strand bg/wig file path')
    parser.add_argument('seqs_file', type=str,
                        help='genome or transcriptome fasta file path')
    parser.add_argument('annotations_file', type=str,
                        help='annotations gtf/gff file path')
    parser.add_argument('-s', '--skip', nargs='+', default=[],
                        help=('list known frameshifted genes that should be '
                        'filtered out before codon assignments'))

    # ORFeus options
    parser.add_argument('--data_file', type=str,
                        help='processed data file path')
    parser.add_argument('-o', '--outdir', type=str, default=os.getcwd(),
                        help=('output directory for all intermediate and '
                        'results files'))

    # Transition probabilities
    parser.add_argument('-a', '--alpha', type=float, default=1e-7,
                        help='probability of PRF')
    parser.add_argument('-b', '--beta', type=float, default=0.5,
                        help='probability of -1 PRF given PRF')
    parser.add_argument('-g', '--gamma', type=float, default=1e-4,
                        help='probability of stop codon readthrough')
    parser.add_argument('-d', '--delta', type=float, default=1e-3,
                        help='probability of upstream or downstream short ORFs')
    parser.add_argument('-z', '--zeta', type=float, default=1e-10,
                        help='probability of multiple non-overlapping ORFs')

    # Feature frequencies
    parser.add_argument('--f5', type=float,
                        help=('fraction of transcripts with 5\'UTRs '
                        '(default: calculated from annotations)'))
    parser.add_argument('--f3', type=float,
                        help=('fraction of transcripts with 3\'UTRs '
                        '(default: calculated from annotations)'))

    # Feature lengths
    parser.add_argument('--utr5', type=int,
                        help=('mean nucleotide length of 5\'UTRs '
                        '(default: calculated from annotations)'))
    parser.add_argument('--utr3', type=int,
                        help=('mean nucleotide length of 3\'UTRs '
                        '(default: calculated from annotations)'))
    parser.add_argument('--orf', type=int,
                        help=('mean nucleotide length of main ORFs '
                        '(default: calculated from annotations)'))
    parser.add_argument('--uorf', type=int, default=50,
                        help=('mean nucleotide length of upstream short ORFs '
                        '(suggested: 50)'))
    parser.add_argument('--dorf', type=int, default=50,
                        help=('mean nucleotide length of downstream short ORFs '
                        '(suggested: 50)'))

    # Feature values
    parser.add_argument('--start_codons', type=str, nargs='*', default=[],
                        help=('list of valid start codons '
                        '(default: calculated from annotations)'))
    parser.add_argument('--stop_codons', type=str, nargs='*', default=[],
                        help=('list of valid stop codons '
                        '(default: calculated from annotations)'))
    parser.add_argument('--sense_codons', type=str, nargs='*', default=[],
                        help=('list of valid sense codons '
                        '(default: calculated from annotations)'))

    # Emission probabilities
    parser.add_argument('-r', '--bins', type=int, default=25,
                        help=('number of bins to group emission probabilities '
                        '(suggested: 25)'))
    parser.add_argument('-l', '--log', default=False, action='store_true',
                        help=('whether to bin the riboseq values in logspace '
                        '(suggested: false)'))
    parser.add_argument('-m', '--min', type=int, default=-1,
                        help=('minimum reads per transcript '
                        '(default: -1)'))
    parser.add_argument('-M', '--max', type=int, default=np.inf,
                        help=('maximum reads per transcript '
                        '(default: infinity)'))
    parser.add_argument('-p', '--pool', default=False, action='store_true',
                        help=('whether to pool the observed riboseq values '
                        'from all codons of a given state type '
                        '(suggested: false, unless you have a small '
                        'transcriptome with <=20 ORFs)'))
    parser.add_argument('-f', '--fit', default=False, action='store_true',
                        help=('whether to fit a log-normal distribution to the '
                        'observed riboseq values or use the raw frequencies '
                        '(suggested: false, unless you have a small '
                        'transcriptome with <=20 ORFs)'))

    # Simulation parameters
    parser.add_argument('-c', '--coverage', default=False, action='store_true',
                        help='whether to run the coverage threshold simulation')
    parser.add_argument('--iters', type=int, default=10,
                        help='number of iterations to simulate')
    parser.add_argument('--window', type=int, default=10,
                        help=('window (in nts) around true event to count as '
                              'correct altORF prediction'))

    args = parser.parse_args()

    print_logo()

# ------------------------------------------------------------------------------
# Prepare output directory
# ------------------------------------------------------------------------------

    # Create output directory
    if args.outdir == os.getcwd():
        outdir = os.path.join(args.outdir, r'orfeus')
        os.makedirs(outdir, exist_ok=True)
    else:
        if not os.path.isdir(args.outdir):
            print('Output directory does not exist. Please provide a valid path.')
        outdir = args.outdir

    # Set up logging
    #log_file = os.path.join(outdir, 'orfeus_build.log')
    logging.basicConfig(format='%(levelname)s: %(message)s',
                        filemode = 'w',
                        level = logging.DEBUG)

# ------------------------------------------------------------------------------
# Process data
# ------------------------------------------------------------------------------

    if args.data_file:
        logging.info('Importing the processed data...')

        # Import the data (specify dtypes where necessary)
        data_df = read_data_file(args.data_file)

        logging.info('Finished importing the data.')

    else:
        logging.info('Processing the data...')

        data_file = os.path.join(outdir, 'data.txt.gz')

        # Check if file already exists
        if os.path.exists(data_file):
            logging.error(('Processed data file already exists. '
                        'If you\'d like to import the data from this file, '
                        'use the --data_file option. '
                        'Otherwise, provide a new output directory path.'))

        # If the file does not exist, create it
        else:

            # Process the data
            data_df = process_data(args.plus_file,
                                   args.minus_file,
                                   args.seqs_file,
                                   args.annotations_file,
                                   skip=args.skip)

            # Export to a tsv file
            data_df.to_csv(data_file, sep='\t', compression='gzip')

            logging.info('Finished processing the data.')

# ------------------------------------------------------------------------------
# Set parameters
# ------------------------------------------------------------------------------

    logging.info('Setting the model parameters...')

    # Set parameters for H1 model (altORFs and canonical ORFs)
    parameters_h1 = set_parameters(data_df,
                                   args.f5,
                                   args.f3,
                                   args.utr5,
                                   args.utr3,
                                   args.orf,
                                   args.uorf,
                                   args.dorf,
                                   args.start_codons,
                                   args.sense_codons,
                                   args.stop_codons,
                                   args.bins,
                                   args.log,
                                   args.min,
                                   args.max,
                                   args.pool,
                                   args.fit,
                    [args.alpha, args.beta, args.gamma, args.delta, args.zeta])

    # Set parameters for H0 model (canonical ORFs only)
    parameters_h0 = set_parameters(data_df,
                                   args.f5,
                                   args.f3,
                                   args.utr5,
                                   args.utr3,
                                   args.orf,
                                   args.uorf,
                                   args.dorf,
                                   args.start_codons,
                                   args.sense_codons,
                                   args.stop_codons,
                                   args.bins,
                                   args.log,
                                   args.min,
                                   args.max,
                                   args.pool,
                                   args.fit,
                                  [0,0,0,0,0])

    # Save the parameters
    np.save(os.path.join(outdir, 'parameters_h1'),
                parameters_h1)
    np.save(os.path.join(outdir, 'parameters_h0'),
                parameters_h0)

    logging.info('Finished setting parameters')

# ------------------------------------------------------------------------------
# Coverage threshold
# ------------------------------------------------------------------------------

    if args.coverage:

        logging.info('Simulating sequences...')

        coverage_outdir = os.path.join(outdir, 'coverage')
        os.makedirs(coverage_outdir, exist_ok=True)

        # Set coverage thresholds for testing
        coverages = np.arange(0.1, 5, 0.1)

        coverage_threshold_simulation(data_df,
                                      parameters_h1,
                                      coverages,
                                      args.iters,
                                      args.window,
                                      coverage_outdir)

        logging.info('Finished simulating sequences')




if __name__== "__main__":
    main()



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# Set emission probabilities
# Author: Mary Richardson
# Date: 2020.07.07, last update 2023.03.02
# -----------------------------------------------------------------------------

import logging
import numpy as np
import pandas as pd
import scipy.stats as ss

logger = logging.getLogger(__name__)

from orfeus.parameters.helper import get_codon, get_nt


def set_nt_emission_probs(states,
                          states_idx,
                          nts):
    """
    Calculate the probability of observing each nucleotide in each state
    Args:
        states: list of all states in the model
        states_idx: dictionary mapping each general state type to its
            state index (1-indexed)
        nts: list of nucleotides
    Returns:
        p_emit_nt: the nucleotide emission probability matrix,
            i.e. if the state is ATG, the prob of A at the first nucleotide is 1
            and the prob of all other nucleotides is 0 (note that this array
            has one row per nucleotide, in the same order as the nts array)
    Raises:
        ValueError: Invalid nt emission probability matrix
    """
    # Prepare the nucleotide emission probabilities matrix
    M = len(states)
    N = len(nts)
    p_emit_nt = np.zeros((M,N))

    # Get a list of noncodon states (can be any nt)
    noncodon_states = [states_idx['5UTR'], states_idx['3UTR']]
    if 'X1' in states:
        noncodon_states.append(states_idx['X1'])
    if 'X2' in states:
        noncodon_states.append(states_idx['X2'])

    # Iterate through row for each possible observed nt
    for i in range(N):

        # For each possible state
        for s in range(M):

            if s in noncodon_states:
                # Any nt could be observed, so set value to 1/4
                p_emit_nt[s, i] = 1/4

            else:
                # Only the nt at this position in this codon could be observed
                codon = get_codon(s, states) # Sequence of this codon
                nt_pos = get_nt(s, states)   # Subcodon nucleotide

                # If this nt is in this state, change the value to 1
                if nts[i] == codon[nt_pos]:
                    p_emit_nt[s, i] = 1

    # Check that emission probabilities for each state sum to 1
    if not np.allclose(np.sum(p_emit_nt, axis=1), np.ones(len(p_emit_nt))):
        logging.error(('Invalid nucleotide emission probability matrix. '
                    'Nucleotide emissions for each state do not sum to 1.'))

    return p_emit_nt


def rho_bins(df, num_bins, logspace=False):
    """
    Fit a distribution to the observed riboseq values for each state
    Args:
        df: pandas DataFrame with all riboseq info
        num_bins: int number of bins to group the riboseq data into
        logspace: boolean indicating whether to bin in logspace
    Returns:
        rho_ranges: array of riboseq emission bin ranges

    """
    # List of ranges for discretizing riboseq observations
    if logspace:
        min_reads = np.nanmin(df['log_reads'])
        max_reads = np.nanmax(df['log_reads'])
    else:
        min_reads = np.nanmin(df['norm_reads'])
        max_reads = np.nanmax(df['norm_reads'])

    # Create R-1 discrete bins for non-zero values
    # (evenly spaced in log or linear space, depending on limits)
    rho_ranges = np.linspace(min_reads, max_reads, num=num_bins)

    # Make sure bin boundaries are specified in linear space
    if logspace:
        rho_ranges = np.exp(rho_ranges)

    # Add a bin for zeros (dummy lower boundary)
    rho_ranges[0] = 0
    rho_ranges = np.insert(rho_ranges, 0, -.0001) # Add a bin for zeros (dummy lower boundary)

    return rho_ranges


def fit_rho_emissions(df,
                      states,
                      states_idx,
                      rho_ranges,
                      pool):
    """
    Fit a distribution to the frequency of observed riboseq values for
        each state
    Args:
        df: pandas DataFrame with all riboseq info
        states: list of state names
        states_idx: dictionary mapping state type to
            state indices (1-indexed)
        rho_ranges: array of riboseq emission bin ranges
        pool: boolean indicating whether to pool all codons of a given type
            (start, stop, sense) when calculating riboseq emissions
    Returns:
        p_emit_rho: the riboseq emission probability matrix
    Raises:

    """
    # Isolate state labels
    df['nt_pos'].fillna(-1, inplace=True)
    df_states = df[['state','codon_type','nt_pos']].drop_duplicates()

    # Update SENSE states (aggregate of all SENSE codon states if pool)
    if pool:
        df_grouped = df.groupby(['codon_type','nt_pos']) \
                       .agg({'log_reads': lambda x: x.dropna().tolist(),
                             'norm_reads': lambda x: x.dropna().tolist()})
        df_grouped = df_states.join(df_grouped,
                                    on=['codon_type','nt_pos'],
                                    how='left')

    # Update START and STOP states
    # (aggregate all START or STOP codon states)
    else:
        df_grouped1 = df.groupby(['state']) \
                        .agg({'log_reads': lambda x: x.dropna().tolist(),
                              'norm_reads': lambda x: x.dropna().tolist()})
        df_grouped2 = df.groupby(['codon_type','nt_pos']) \
                        .agg({'log_reads': lambda x: x.dropna().tolist(),
                              'norm_reads': lambda x: x.dropna().tolist()})
        df_grouped1 = df_states.join(df_grouped1,
                                     on='state',
                                     how='left')
        df_grouped2 = df_states.join(df_grouped2,
                                     on=['codon_type','nt_pos'],
                                     how='left')

        df_grouped1 = df_grouped1[~df_grouped1['codon_type'] \
                                 .isin(['START','STOP'])]
        df_grouped2 = df_grouped2[df_grouped2['codon_type'] \
                                 .isin(['START','STOP'])]
        df_grouped = pd.concat([df_grouped1, df_grouped2])

    # Fit a log normal distribution to the nonzero reads
    df_grouped['p'] = df_grouped.apply(lambda row: \
                                 ss.norm.fit(row['log_reads']),
                                 axis=1)

    # Calculate the the fraction of zero reads
    df_grouped['q'] = df_grouped.apply(lambda row: \
                                 float((len(row['norm_reads']) -
                                 np.count_nonzero(row['norm_reads'])) /
                                 len(row['norm_reads'])),
                                 axis=1)

    # Make sure all states are in order
    df_grouped.set_index('state', inplace=True)
    df_grouped = df_grouped.reindex(states)

    # Keep only these parameters for each state
    p = df_grouped['p'].tolist()
    q = df_grouped['q'].tolist()

    # Set representative emission values (mean each bin greater than zero)
    emit = np.mean([rho_ranges[:-1], rho_ranges[1:]], axis=0)

    # Discretize the emissions
    M = len(states)
    R = len(rho_ranges)-1
    p_emit_rho = np.zeros([M,R])

    # Skip altORF states for now
    for i in (states_idx['5UTR'] + states_idx['ORF'] + states_idx['3UTR']):

        # Add the zero emissions
        if np.any(np.isnan(q[i])):
            p_emit_rho[i,0] = None
        else:
            p_emit_rho[i,0] = q[i]

        # Add the nonzero emissions
        for j in range(len(emit)-1):
            if np.any(np.isnan(p[i])):
                p_emit_rho[i,j+1] = None
            else:
                p_emit_rho[i,j+1] = ss.norm(*p[i]).pdf(emit[j]) * (1-q[i])

    return p_emit_rho


def nofit_rho_emissions(df,
                        states,
                        states_idx,
                        rho_ranges,
                        pool):
    """
    Calculate the frequency observed riboseq values for each state
    Args:
        df: pandas DataFrame with all riboseq info
        states: list of state names
        states_idx: dictionary mapping state type to
            state indices (1-indexed)
        rho_ranges: array of riboseq emission bin ranges
        pool: boolean indicating whether to pool all codons of a given type
            (start, stop, sense) when calculating riboseq emissions
    Returns:
        p_emit_rho: the riboseq emission probability matrix
    Raises:

    """# Group the log riboseq reads by range
    # (include the upper limit of the range)
    df['rho_ranges'] = pd.cut(df['norm_reads'],
                                    rho_ranges,
                                    right=True)

    # Get an index of all states and log read ranges
    df_states = list(set(df['state']) & set(states))
    idx = pd.MultiIndex \
            .from_product([df_states, df['rho_ranges'].cat.categories],
                          names=['state', 'rho_ranges'])

    # For each state, count the reads in each range
    df = df.groupby(['state', 'rho_ranges']) \
           .size()

    # Reset the row names to make sure ALL log reads ranges are included
    # (fill in zeros for any missing ranges)
    df = df.reindex(index=idx, fill_value=0) \
           .reset_index()
    df.columns = ['state', 'rho_ranges', 'counts'] # Rename columns

    # Pivot to a MxR matrix of counts for each state
    df = df.pivot(index='state',
                  columns='rho_ranges',
                  values='counts')

    # Reorder the states
    df = df.reindex(states)

    # Emission prob is the frequency of counts in each state
    p_emit_rho = df.to_numpy()

    # Update SENSE states (mean of all SENSE codon states if pool)
    if pool:
        for nt in [0,1,2]:
            mean_sense = \
                np.nanmean(p_emit_rho[nt+states_idx['sense'][0]:states_idx['sense'][-1]+1:3],
                        axis=0)
            p_emit_rho[nt+states_idx['sense'][0]:states_idx['sense'][-1]+1:3] \
                = mean_sense

    return p_emit_rho


def set_rho_emission_probs(data_df,
                           states,
                           states_idx,
                           num_bins,
                           logspace,
                           min_coverage,
                           max_coverage,
                           pool,
                           fit):
    """
    Calculate the probability of observing observing each range of riboseq
        values in each state
    Args:
        data_df: pandas DataFrame with all transcript info
        states: list of all states in the model
        states_idx: dictionary mapping each general state type to its
            state index (1-indexed)
        num_bins: int number of bins to group the riboseq data into
        logspace: (optional) boolean indicating whether to bin in logspace
        min_coverage: (optional) int minimum mean reads per transcript
        max_coverage: (optional) int maximum mean reads per transcript
        pool: (optional) boolean indicating whether to pool all codons of a given type
            (start, stop, sense) when calculating riboseq emissions
        fit: (optional) boolean indicating whether to fit a log-normal distribution to
            the observed non-zero values when calculating riboseq emissions
    Returns:
        p_emit_rho: the riboseq emission probability matrix
        rho_ranges: array of riboseq emission bin ranges
    Raises:
        ValueError: Invalid riboseq emission probability matrix
    """
    # Filter down to states that are in the list of states
    df = data_df[data_df['state'].isin(states)]

    # Filter down to transcripts above min coverage and below max coverage
    df = df[(df['mean_reads']>=min_coverage) & (df['mean_reads']<=max_coverage)]

    # Generate riboseq emission bin ranges
    rho_ranges = rho_bins(df, num_bins, logspace)

    # Calculate the emission probabilities from the observed riboseq data
    # for all canonical states
    if fit:
        p_emit_rho = fit_rho_emissions(df,
                                       states,
                                       states_idx,
                                       rho_ranges,
                                       pool)
    else:
        p_emit_rho = nofit_rho_emissions(df,
                                       states,
                                       states_idx,
                                       rho_ranges,
                                       pool)

    # Update START states (mean of all START codon states)
    for nt in [0,1,2]:
        mean_start = np.nanmean(p_emit_rho[nt+states_idx['start'][0]:states_idx['start'][-1]+1:3], \
                              axis=0)
        p_emit_rho[nt+states_idx['start'][0]:states_idx['start'][-1]+1:3] \
                = mean_start
        if 'start_uORF' in states_idx:
            p_emit_rho[nt+states_idx['start_uORF'][0]:states_idx['start_uORF'][-1]+1:3] \
                = mean_start
        if 'start_dORF' in states_idx:
            p_emit_rho[nt+states_idx['start_dORF'][0]:states_idx['start_dORF'][-1]+1:3] \
                = mean_start

    # Update STOP states (mean of all STOP codon states)
    for nt in [0,1,2]:
        mean_stop = np.nanmean(p_emit_rho[nt+states_idx['stop'][0]:states_idx['stop'][-1]+1:3], \
                             axis=0)
        p_emit_rho[nt+states_idx['stop'][0]:states_idx['stop'][-1]+1:3] \
                = mean_stop
        if 'stop_uORF' in states_idx:
            p_emit_rho[nt+states_idx['stop_uORF'][0]:states_idx['stop_uORF'][-1]+1:3] \
                = mean_stop
        if 'stop_dORF' in states_idx:
            p_emit_rho[nt+states_idx['stop_dORF'][0]:states_idx['stop_dORF'][-1]+1:3] \
                = mean_stop
        if 'stop_SCR' in states_idx:
            p_emit_rho[nt+states_idx['stop_SCR'][0]:states_idx['stop_SCR'][-1]+1:3] \
                = mean_stop

    # Add uORF and dORF states
    if 'sense_uORF' in states_idx:
        p_emit_rho[states_idx['sense_uORF'][0]:states_idx['sense_uORF'][-1]+1] = \
            p_emit_rho[states_idx['sense'][0]:states_idx['sense'][-1]+1]
    if 'sense_dORF' in states_idx:
        p_emit_rho[states_idx['sense_dORF'][0]:states_idx['sense_dORF'][-1]+1] = \
        p_emit_rho[states_idx['sense'][0]:states_idx['sense'][-1]+1]

    # Add PRF states
    mean_orf = np.nanmean(p_emit_rho[states_idx['sense'][0]:states_idx['sense'][-1]+1],
                          axis=0)
    if 'X1' in states:
        p_emit_rho[states_idx['X1']] = mean_orf # X1
        if 'X2' in states:
            p_emit_rho[states_idx['X2']] = mean_orf # X2

    # Replace missing states
    for row in np.where(~p_emit_rho.any(axis=1))[0]:
        logging.warning('Emissions missing for ' + states[row])
        p_emit_rho[row] = mean_orf

    # Add a pseudocount to all counts
    p_emit_rho += 1

    # Normalize to make sure the probs for each state sum to 1
    p_emit_rho = p_emit_rho / p_emit_rho.sum(axis=1, keepdims=True)

    # Check that emission probabilities for each state sum to 1
    if not np.allclose(np.sum(p_emit_rho, axis=1), np.ones(len(p_emit_rho))):
        logging.error(('Invalid riboseq emission probability matrix. '
                        'Riboseq emissions for each state do not sum to 1.'))

    return p_emit_rho, rho_ranges



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# Set initial transition and emission probs
# Author: Mary Richardson
# Date: 2020.07.07, last update 2023.03.02
# -----------------------------------------------------------------------------

import numpy as np
import pandas as pd
import itertools as it

from orfeus.parameters.states import *
from orfeus.parameters.transitions import *
from orfeus.parameters.emissions import *


NUCLEOTIDES = ['A','C','G','T']
CODONS = [''.join(i) for i in it.product(NUCLEOTIDES, repeat = 3)]


def set_parameters(data_df,
                   f_5UTR,
                   f_3UTR,
                   len_5UTR,
                   len_3UTR,
                   len_orf,
                   len_uorf,
                   len_dorf,
                   start_codons,
                   sense_codons,
                   stop_codons,
                   num_bins,
                   logspace,
                   min_coverage,
                   max_coverage,
                   pool,
                   fit,
                   parameters):
    """
    Set the model transition and emission probabilities
    Args:
        data_df: pandas DataFrame with all transcript info
        f_5UTR: float indicating the fraction of ORFs having 5'UTRs
        f_3UTR: float indicating the fraction of ORFs having 3'UTRs
        len_5UTR: float indicating the mean length of 5'UTRs
        len_3UTR: float indicating the mean length of 3'UTRs
        len_orf: float indicating the mean length of ORFs
        len_uorf: float indicating the mean length of uORFs
        len_dorf: float indicating the mean length of dORFs
        start_codons: list of valid start codons
        sense_codons: list of valid sense codons
        stop_codons: list of valid stop codons
        num_bins: int number of bins to group the riboseq data into
        logspace: boolean indicating whether to bin in logspace
        min_coverage: int minimum mean reads per transcript
        max_coverage: int maximum mean reads per transcript
        pool: boolean indicating whether to pool all codons of a given type
            (start, stop, sense) when calculating riboseq emissions
        fit: boolean indicating whether to fit a log-normal distribution to
            the observed non-zero values when calculating riboseq emissions
        parameters: list of parameter values for
            alpha, beta, gamma, delta, zeta
    Returns:
        states: list of state names
        states_idx: dictionary mapping state type to
            state indices (1-indexed)
        altorfs_idx: dictionary mapping more general state type to
            state indices (0-indexed)
        p_emit_nt: the nucleotide emission probability matrix
        p_emit_rho: the riboseq emission probability matrix
        p_trans: the transition probability matrix
        p_trans_rev: the reverse transition probability matrix
        prev_state_options: list of indices of all possible
            (nonzero probability) previous states
        next_state_options: list of indices of all possible
            (nonzero probability) next states
    """
    # Only train on annotated regions of transcripts
    data_df.drop(data_df[~data_df['feature'] \
           .isin(['5UTR', 'ORF', '3UTR'])].index, inplace=True)

    # Set constants
    f_5UTR, f_3UTR, len_5UTR, len_3UTR, len_orf, len_uorf, len_dorf, \
    start_codon_freqs, sense_codon_freqs, stop_codon_freqs = \
        set_constants(data_df,
                    CODONS,
                    f_5UTR,
                    f_3UTR,
                    len_5UTR,
                    len_3UTR,
                    len_orf,
                    len_uorf,
                    len_dorf,
                    start_codons,
                    sense_codons,
                    stop_codons)

    # Set model states
    states, states_idx, altorfs_idx = \
        set_states(start_codon_freqs,
                    sense_codon_freqs,
                    stop_codon_freqs,
                    parameters)

    # Set emission probabilities
    p_emit_nt = \
        set_nt_emission_probs(states,
                    states_idx,
                    NUCLEOTIDES)

    p_emit_rho, rho_ranges = \
        set_rho_emission_probs(data_df,
                    states,
                    states_idx,
                    num_bins,
                    logspace,
                    min_coverage,
                    max_coverage,
                    pool,
                    fit)

    # Set transition probabilities
    p_trans = \
        set_transition_probs(states,
                    states_idx,
                    f_5UTR,
                    f_3UTR,
                    len_3UTR,
                    len_5UTR,
                    len_orf,
                    len_uorf,
                    len_dorf,
                    start_codon_freqs,
                    sense_codon_freqs,
                    stop_codon_freqs,
                    parameters)

    p_trans_rev = rev_transition_probs(p_trans)

    prev_states_options = get_prev_states(p_trans)
    next_states_options = get_next_states(p_trans)

    all_parameters = np.array([states, states_idx, altorfs_idx, \
           p_emit_nt, p_emit_rho, rho_ranges, p_trans, p_trans_rev, \
           prev_states_options, next_states_options], dtype=object)

    return all_parameters



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# HMM output functions
# Author: Mary Richardson
# Date: 2021.10.25
# -----------------------------------------------------------------------------

import logging
import numpy as np
import itertools as it
import multiprocessing as mp

logger = logging.getLogger(__name__)


AMINO_ACIDS = {'TTT':'F','TTC':'F',
               'TTA':'L','TTG':'L',
               'TCT':'S','TCC':'S','TCA':'S','TCG':'S','AGT':'S','AGC':'S',
               'TAT':'Y','TAC':'Y',
               'TAA':'*','TAG':'*','TGA':'*',
               'TGT':'C','TGC':'C',
               'TGG':'W',
               'CTT':'L','CTC':'L','CTA':'L','CTG':'L',
               'CCT':'P','CCC':'P','CCA':'P','CCG':'P',
               'CAT':'H','CAC':'H',
               'CAA':'Q','CAG':'Q',
               'CGT':'R','CGC':'R','CGA':'R','CGG':'R','AGA':'R','AGG':'R',
               'ATT':'I','ATC':'I','ATA':'I',
               'ATG':'M',
               'ACT':'T','ACC':'T','ACA':'T','ACG':'T',
               'AAT':'N','AAC':'N',
               'AAA':'K','AAG':'K',
               'GTT':'V','GTC':'V','GTA':'V','GTG':'V',
               'GCT':'A','GCC':'A','GCA':'A','GCG':'A',
               'GAT':'D','GAC':'D',
               'GAA':'E','GAG':'E',
               'GGT':'G','GGC':'G','GGA':'G','GGG':'G'}


def get_ranges(values):
    """
    Code source: https://stackoverflow.com/questions/4628333/converting-a-list-of-integers-into-range-in-python
    Yields ranges of values present in a list
    (e.g. [2,3,4,8,9] would give ranges 2-4 and 8-9)
    Args:
        values: list of integer values
    Returns:

    Raises:

    """
    for a, b in it.groupby(enumerate(values), lambda pair: pair[1] - pair[0]):
        b = list(b)
        yield b[0][1], b[-1][1]


def get_events(path, altorfs_idx):
    """
    Finds all altORF events and their indices in the given path
    Args:
        path: the state path
        altorfs_idx: dictionary mapping more general state type to
            state indices (0-indexed)
    Returns:
        altorfs: list of all ORF and altORF events and corresponding indices
            in the path, in string format for printing
    Raises:

    """
    altorfs = []

    if len(path)==0:
        return altorfs

    for event in ['PRF','SCR','uORF','dORF','ORF']:
        if event in altorfs_idx:

            event_idx = set(altorfs_idx[event])
            path_idx = [i for i,s in enumerate(path) if s in event_idx]

            ranges = list(get_ranges(path_idx))
            for range in ranges:

                # Print single position
                if event=='PRF':
                    if len(range)==1: # +1 PRF
                        altorfs.append('+1 %s (%i)' % (event, range[0]))
                    elif len(range)==2: # -1 PRF
                        altorfs.append('-1 %s (%i)' % (event, range[0]))

                # Print single position
                if event=='SCR':
                    altorfs.append('%s (%i)' % (event, range[1]))

                # Print interval of positions
                else:
                    altorfs.append('%s (%i-%i)' % (event, range[0], range[1]))

    return altorfs


def readable_sequence(x,
                      path,
                      states_idx):
    """
    Generates a human-readable representation of the predicted state path for
    the given nucleotide sequence:
        [] = bounds of main ORFs, uORFs, dORFs
         | = SCR site
        () = PRF skipped nucleotide (+1 PRF) or repeated nucleotides (-1 PRF)
         . = UTR nucleotide
    Args:
        x: the nucleotide sequence
        path: the state path
        states_idx: dictionary mapping each general state type to its
            state index (1-indexed)
    Returns:
        readable_seq: the state path in a readable format
    Raises:

    """
    readable_seq = ''

    for i in range(len(x)):

        # UTR
        if path[i]==states_idx['5UTR'] or path[i]==states_idx['3UTR']:
            readable_seq += '.'

        # uORF start codon (1st nt)
        elif path[i] in range(states_idx['start_uORF'][0],
                              states_idx['start_uORF'][-1],
                              3):
            readable_seq += '[' + x[i]

        # dORF start codon (1st nt)
        elif path[i] in range(states_idx['start_dORF'][0],
                              states_idx['start_dORF'][-1],
                              3):
            readable_seq += '[' + x[i]

        # ORF start codon (1st nt)
        elif path[i] in range(states_idx['start'][0],
                              states_idx['start'][-1],
                              3):
            readable_seq += '[' + x[i]

        # uORF stop codon (3rd nt)
        elif path[i] in range(states_idx['stop_uORF'][0]+2,
                              states_idx['stop_uORF'][-1],
                              3):
            readable_seq += x[i] + ']'

        # dORF stop codon (3rd nt)
        elif path[i] in range(states_idx['stop_dORF'][0]+2,
                              states_idx['stop_dORF'][-1],
                              3):
            readable_seq += x[i] + ']'

        # ORF stop codon (3rd nt)
        elif path[i] in range(states_idx['stop'][0]+2,
                              states_idx['stop'][-1],
                              3):
            readable_seq += x[i] + ']'

        # SCR
        elif (path[i] in range(states_idx['stop_SCR'][0]+2,
                               states_idx['stop_SCR'][-1],
                               3)):
            readable_seq += x[i] + '|'

        # PRF
        elif path[i]==states_idx['X1'] or path[i]==states_idx['X2']:
            readable_seq += '(' + x[i] + ')'

        # Sense codon
        else:
            readable_seq += x[i]

    return readable_seq


def get_orfs(readable_seq):
    """
    Extracts the nucleotide sequence for each ORF from a prediction result
    Args:
        readable_seq: the state path in a readable format
    Returns:
        seqs: nucleotide sequences for each ORF
    Raises:

    """
    print(readable_seq)
    # Get the sequence and info for each ORF
    seqs = []
    for x in readable_seq:
        if x=='.':
            continue
        elif x=='[':
            seq = ''
        elif x==']':
            seqs.append(seq)
        elif x=='(':
            continue
        elif x==')': # Skip the frameshift nucleotide
            seq = seq[:-1] # (really it could be repeated for a -1 or -2 PRF)
        elif x=='|': # Skip the readthrough stop
            seq = seq[:-3] # (really it could be replaced by a near-cognate codon)
        else:
            seq += x

    return seqs


def translate(x, amino_acids):
    """
    Translate a nucleotide sequence to protein
    Args:
        x: nucleotide sequence
        amino_acids: dict mapping each codon to an amino acid (codon table)
    Returns:
        y: protein sequence
    Raises:

    """
    # Translate the nt sequence using the codon table
    L = len(x)
    y = [amino_acids[''.join(x[i:i+3])] for i in range(0,L-2,3)]

    # If there is no stop codon return an empty sequence
    if '*' not in y:
        logging.warning('Couldn\'t translate sequence (missing stop codon): %s' % x)
        return ''

    # If there is a stop codon, truncate the protein sequence
    else:
        # Find the first stop codon
        idx_stop = y.index('*')

        # Terminate the protein sequence right before the first stop codon
        return ''.join(y[:idx_stop])


def trace_dump(x,
               r,
               path,
               states,
               log_emit,
               log_trans):
    """
    Prints the path score at each position of the sequence
    Args:
        x: the nucleotide sequence
        r: the rho sequence
        path: the state path
        states: list of all states in the model
        log_emit: the total emission probabilities at each position for
            this sequence (combined nucleotide and rho emissions)
        log_trans: log transition probability matrix
    Returns:

    Raises:

    """
    # Length of observed sequence
    L = len(path)

    if L==0:
        return np.nan

    # Print the header
    print('{:5s} {:5s} {:15s} {:5s} {:10s} {:15s} {:15s} {:15s}'\
          .format('pos',
                  'nt',
                  'rho',
                  'state_idx',
                  'state',
                  'log_transition',
                  'log_emission',
                  'log_posterior'))

    # Add the probability of the initial transition
    # from the begin state (in log space)
    P = log_trans[0, path[0]+1]

    # For each position in the sequence 0..L-1
    # add the probability of the transition to this residue and
    # the emission for this residue (in log space)
    for i in range(0, L-1):
        P += log_emit[i, path[i]] + log_trans[path[i]+1, path[i+1]+1]
        print('{:5.0f} {:5s} {:15.10f} {:5.0f} {:10s} {:15.10f} {:15.10f} {:15.10f}'\
              .format(i,
                      x[i],
                      r[i],
                      path[i],
                      states[path[i]],
                      log_trans[path[i]+1, path[i+1]+1],
                      log_emit[i, path[i]],
                      P))

    # Add the probability of the final emission and
    # the transition to the end state (in log space)
    P += log_emit[L-1, path[L-1]] + log_trans[path[L-1]+1, 0]
    print('{:5.0f} {:5s} {:15.10f} {:5.0f} {:10s} {:15.10f} {:15.10f} {:15.10f}'\
          .format(L-1,
                  x[L-1],
                  r[L-1],
                  path[L-1],
                  states[path[L-1]],
                  log_trans[path[i]+1,
                  path[i+1]+1],
                  log_emit[i, path[i]],
                  P))


def write_results_fasta(results,
                        parameters_h1,
                        fn_file,
                        fa_file,
                        chunk=60):
    """
    Writes a fasta file of predicted nucleotide and protein sequences
    Args:
        results: array containing all results of running the model
        parameters_h1: list of all parameter matrices for the current model
        file: path to output fasta file
        fn_file: path to output nucleotide sequence fasta file
        fa_file: path to output protein sequence fasta file
        chunk: max line length for sequence entries, add line breaks to
            maintain this max length
    Returns:

    Raises:
        IOError: an error occurred accessing the file
    """
    # Unpack the parameters
    states, states_idx, altorfs_idx, \
    p_emit_nt, p_emit_rho, rho_ranges, p_trans, p_trans_rev, \
    prev_states_options, next_states_options = parameters_h1

    # Open the fasta files for writing
    try:
        fn = open(fn_file,'w')
    except IOError:
        logging.error('Could not open fasta file %s for writing' % file)

    try:
        fa = open(fa_file,'w')
    except IOError:
        logging.error('Could not open fasta file %s for writing' % file)

    else:

        with fn, fa:

            for result in results:

                # Unpack the transcript attributes
                transcript_id, transcript_name, coverage, x, r, \
                          path_h2, path_h1, path_h0, \
                          path_score_h2, path_score_h1, path_score_h0, \
                          log_odds_score = result

                # Write the nucleotide sequence for this transcript
                id = '%s (%s)' % (transcript_id, transcript_name)
                seq = readable_sequence(x, path_h1, states_idx)

                # Write the seq id
                fn.write('>%s orfeus prediction\n' % id)

                # Write the sequence with a fixed max line length
                i = 0
                while i < len(seq)//chunk:
                    fn.write('%s\n' % seq[i*chunk:(i+1)*chunk])
                    i += 1
                fn.write('%s\n' % seq[i*chunk:])

                # Write the protein sequence(s) for this transcript
                xs = get_orfs(seq)
                k = 0
                for x in xs:
                    y = translate(x, AMINO_ACIDS)

                    # Write the seq id
                    fa.write('>%s orfeus prediction %i\n' % (id,k))
                    k += 1

                    # Write the sequence with a fixed max line length
                    i = 0
                    while i < len(y)//chunk:
                        fa.write('%s\n' % y[i*chunk:(i+1)*chunk])
                        i += 1
                    fa.write('%s\n' % y[i*chunk:])


def write_results_table(results,
                        parameters_h1,
                        file):
    """
    Writes a text file of predicted ORF and altORF events
    Args:
        results: array containing all results of running the model
        parameters_h1: list of all parameter matrices for the current model
        file: path to output fasta file
    Returns:

    Raises:
        IOError: an error occurred accessing the file
    """
    # Unpack the parameters
    states, states_idx, altorfs_idx, \
    p_emit_nt, p_emit_rho, rho_ranges, p_trans, p_trans_rev, \
    prev_states_options, next_states_options = parameters_h1

    # Open the fasta file for writing
    try:
        f = open(file,'w')

    except IOError:
        logging.error('Could not open output file %s for writing' % file)

    else:

        # Write the header
        f.write('{:20s} {:20s} {:20s} {:20s} {:20s} {:20s} {:20s} {:20s} {:20s} {:20s}\n' \
                 .format('transcript_id',
                         'transcript_name',
                         'mean_coverage',
                         'predicted=annotated',
                         'annotated_events',
                         'predicted_events',
                         'log_odds_score',
                         'annotated_path_score',
                         'predicted_path_score',
                         'null_path_score'))

        # Sort by log-odds score
        results = sorted(results, key=lambda x: x[-1], reverse=True)

        with f:

            for result in results:

                # Unpack the transcript attributes
                transcript_id, transcript_name, coverage, x, r, \
                          path_h2, path_h1, path_h0, \
                          path_score_h2, path_score_h1, path_score_h0, \
                          log_odds_score = result

                # Get the events
                events_h2 = ', '.join(get_events(path_h2, altorfs_idx))
                events_h1 = ', '.join(get_events(path_h1, altorfs_idx))

                # Check whether annotated ORF = predicted ORF
                if path_score_h2==path_score_h1: canonical = 'true'
                else: canonical = 'false'

                # Write the results
                f.write('{:20s} {:20s} {:20.4f} {:20s} {:20s} {:20s} {:20.3f} {:20.3f} {:20.3f} {:20.3f}\n' \
                        .format(transcript_id,
                                transcript_name,
                                coverage,
                                canonical,
                                events_h2,
                                events_h1,
                                log_odds_score,
                                path_score_h2,
                                path_score_h1,
                                path_score_h0))
